"""
Copyright 2019, University of Freiburg.

Niklas Schnelle <schnelle@cs.uni-freiburg.de>
"""
from typing import List, Optional, Dict, Any, Set, DefaultDict, NamedTuple, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
from abc import ABC, abstractmethod
import queue, sys, threading
import template_handling
from copy import deepcopy
import time


class BackendError(Exception):
    """
    An unspecified backend errror
    TODO(schnelle) better errors
    """

class QueryTime:
    def __init__(self):
        self.client_time = None
        self.query_time = None
        self.compute_time = None
        self.certifyTargetWorked = True
        self.certifyFirstPageWorked = True

class Backend(ABC):
    """
    Backend abstracts a SPARQL endpoint
    """
    def __init__(self):
        self.last_query_client_time = None  # type: Optional[float]
        self.last_query_compute_time = None  # type: Optional[float]
        self.last_query_time = None  # type: Optional[float]
        self.is_qlever = False

    @abstractmethod
    def query(self, query: str, **kwargs) \
            -> T[Optional[List[List[str]]], Dict[str, Any], QueryTime]:
        """Returns the result table of the query as a list of rows.

        :param query:
        :return:
        """

    # !/usr/bin/env python


class ExThread(threading.Thread):
    def __init__(self, f):
        threading.Thread.__init__(self)
        self.__status_queue = queue.Queue()
        self.f = f
        self.result = None

    def run(self):
        """This method should NOT be overriden."""
        try:
            self.result = self.f()
        except BaseException:
            self.__status_queue.put(sys.exc_info())
        self.__status_queue.put(None)

    def wait_for_exc_info(self):
        return self.__status_queue.get()

    def join_with_exception(self):
        ex_info = self.wait_for_exc_info()
        if ex_info is None:
            return self.result
        else:
            raise ex_info[1]

def fallback_with_certificate (backend, fallback_backend, fallback_query, certify_template, num_certificates, certify_target, certify_timeout = 3):
    res, rti, times = fallback_backend.query(fallback_query)
    certify_timeout -= times.query_time
    certify_timeout = max(0.1, certify_timeout) # avoid oddities
    if certify_template is None or not res:
        return res, rti, times

    ts = []
    target_was_found = False
    certified_candidates = []
    certify_queries = []
    def append_certify_query(target):
        certified_candidates.append(target)
        repl = {"CANDIDATE": target}
        q = deepcopy(certify_template)
        q= q.replace("ql:autocomplete_certify_candidate", target)
        certify_queries.append(q)
        t = ExThread(lambda: backend.query(q, timeout=certify_timeout, silent=True, ignore_timeout=False))
        # TODO for debugging
        t.start()
        ts.append(t)
    for i in range(min(num_certificates, len(res))):
        append_certify_query(res[i][0])
        if (res[i][0] == certify_target):
            target_was_found = True
    # also always try to certify the target
    if (not target_was_found):
        append_certify_query(certify_target)
    max_query_time = 0
    num_failed_certificates = 0
    for i, t in enumerate(ts):
        #TODO for debuggin
        #t.start()
        query_res, rti2, t2 = t.join_with_exception()

        if certified_candidates[i] == certify_target:
            # hack the certificate queries into the output, the agnostic queries are not that interesting
            rti = rti2
            if backend.is_qlever:
                try:
                    print("Certificate count of target was {}".format(query_res[0][0]))
                    if query_res[0][0] == 0 or query_res[0][0] == "0":
                        print("Certificate count of target should be higher than 0, certificate query was")
                        print(certify_queries[i])
                        print("Runtime info was")
                        print(rti2)
                except:
                    print("Could not extract certificate count from {}".format(query_res))
                    print("runtime-info was")
                    print(rti2)
                    print("certificate_query was")
                    print(certify_queries[i])



        if t2.query_time is not None :
            max_query_time = max(max_query_time, t2.query_time)
        else:
            # at least one of the certificates timed out
            max_query_time = max(max_query_time, certify_timeout)
        if t2.query_time is None or query_res is None:
            num_failed_certificates += 1
            # this certificate query failed
            if certified_candidates[i] == certify_target:
                times.certifyTargetWorked = False
            if i < min(num_certificates, len(res)) :
                times.certifyFirstPageWorked = False

    if max_query_time is not None and times.query_time is not None:
        times.query_time += max_query_time
    else:
        times.query_time = None

    #time.sleep(1.5)
    print("{} out of {} certificates failed".format(num_failed_certificates, len(certified_candidates)))
    return res, rti, times

def query_async_with_timeout(backend, fallback_backend, query, fallback_query, timeout, certify_template=None, certify_target=None, num_certificates=7):

    q1 = lambda : backend.query(query, timeout=timeout)
    q2 = lambda : fallback_with_certificate(backend, fallback_backend, fallback_query, certify_template, num_certificates, certify_target, timeout)
    t1 = ExThread(q1)
    t2 = ExThread(q2)
    t1.start()
    # TODO for debugging
    t2.start()

    res1 = t1.join_with_exception()
    if res1[0] is None:
        print("No results in sensitive query")
    a1, b1, c1 = res1
    # TODO for debugging
    #t2.start()
    res2 = t2.join_with_exception()
    a2, b2, c2 = res2
    return [a1, a2], [b1, b2], [c1, c2]
