#!/usr/bin/env python3
"""
 Implements a SPARQL backend using the Blazegraph API

 Copyright (c) 2019 University of Freiburg
 Chair of Algorithms and Data Structures
 Author: Niklas Schnelle

"""

from typing import List, Optional, Dict, Any, Set, DefaultDict, NamedTuple, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
import json
import time
import re
import logging
from urllib3 import HTTPConnectionPool, HTTPSConnectionPool, Retry, make_headers
from urllib3.util import Timeout
from urllib3.exceptions import MaxRetryError
import certifi

from . import Backend, BackendError, QueryTime



class Blazegraph(Backend):
    """
    Backend abstracts the Blazegraph SPARQL endpoint
    """
    def __init__(self, backend_host,
                 backend_port,
                 backend_url='/',
                 backend_timeout_seconds=30,
                 clear_cache=False, ssl=False):
        self.is_blazegraph = True #distinguish between blazegraph and virtuoso for the hacks
        super().__init__()
        self.backend_host = backend_host
        self.backend_port = backend_port
        self.backend_url = backend_url
        self.backend_timeout_seconds = backend_timeout_seconds
        self.clear_cache = clear_cache
        self.connection_pool = None  # type: HTTPConnectionPool
        self.ssl = ssl
        self._init_connection_pool(4)

    def _init_connection_pool(self, pool_maxsize):
        # Retry on 404 and 503 messages because
        # these seem to happen sometimes, but very rarely.
        retry = Retry(total=0, status_forcelist=[404, 503],
                      backoff_factor=0.1)
        self.connection_pool = self.create_connection_pool(pool_maxsize, 10)

    def create_connection_pool(self, pool_maxsize, timeout):
        # Retry on 404 and 503 messages because
        # these seem to happen sometimes, but very rarely.
        retry = Retry(total=0, status_forcelist=[404, 503],
                      backoff_factor=0.1)
        timeout *= 1.0
        timeout = Timeout(connect=timeout, read=timeout)
        if self.ssl:
            connection_pool = HTTPSConnectionPool(self.backend_host,
                                                  port=self.backend_port,
                                                  maxsize=pool_maxsize,
                                                  timeout=timeout,
                                                  retries=retry,
                                                  cert_reqs='CERT_REQUIRED',
                                                  ca_certs=certifi.where())
        else:
            connection_pool = HTTPConnectionPool(self.backend_host,
                                                 port=self.backend_port,
                                                 maxsize=pool_maxsize,
                                                 timeout=timeout,
                                                 retries=retry)
        return connection_pool
    def clear_server_cache(self, clearPinned=False) -> bool:
        """
        Instructs the backend server to clear its cache
        """
        return True


    def query(self, query: str, **kwargs) \
            -> T[Optional[List[List[str]]], Dict[str, Any], QueryTime]:
        """Returns the result table of the query as a list of rows.

        :param query:
        :return:
        """

        timeout = kwargs.get('timeout', self.backend_timeout_seconds)
        pool = self.connection_pool if timeout is None else self.create_connection_pool(2, timeout)

        # by default do not send the timeout to virtuoso since it seems to benefit from this behavior
        ignore_timeout = kwargs.get("ignore_timeout", True)

        # this hack is required for blazegraph in the freebase-easy mode
        if self.is_blazegraph:
            query = re.sub(r"PREFIX\s+fbe:\s*<>", "PREFIX fbe: <file:/local/data/blazegraph/freebase-easy/>", query)
            # sometimes we now have the fbe: prefix twice
            query = re.sub("PREFIX fbe: <file:/local/data/blazegraph/freebase-easy/>\s+PREFIX fbe: <file:/local/data/blazegraph/freebase-easy/>","PREFIX fbe: <file:/local/data/blazegraph/freebase-easy/>", query)

        if self.is_blazegraph:
            params = {
                "query": query,
                # "timeout": 5,
                "maxQueryTimeMillis": 1000 * timeout
            }
        else:
            params = {
                "query": query,
                 "timeout": 1000 * timeout,
            }

            if not ignore_timeout:
                params["timeout"] = 1000 * timeout
        # The new blazegraph api will send us 403 if we don't pretend to be some
        # valid user agent
        fake_browser = {'User-Agent': 'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/41.0.2228.0 Safari/537.3'}

        
        headers = {
                'Accept': 'application/sparql-results+json',
                'Keep-Alive': True}
        headers.update(fake_browser)
        table = None  # type: Optional[List[List[str]]]
        runtime_info = {}  # type: Dict[str, Any]
        start = time.time()
        times = QueryTime()
        try:
            resp = pool.request('GET',
                                                self.backend_url,
                                                headers=headers,
                                                fields=params)
            times.query_time = (time.time() - start)
            times.client_time = times.query_time

            if resp.status == 200:
                table = []
                data = json.loads(resp.data.decode('utf-8'))
                if 'results' not in data:
                    logging.warning('No query results in backend answer')
                    times.query_time = min(timeout, times.client_time)
                    return None, runtime_info, times
                col_map = {key: i for i, key in enumerate(data['head']['vars'])}
                results = data['results']
                bindings = results['bindings']
                for binding in bindings:
                    row = [''] * len(col_map)
                    for var, obj in binding.items():
                        value = '"' + obj['value'] + '"'
                        if obj['type'] == 'uri':
                            value = '<' + value.strip('"') + '>'
                        elif 'datatype' in obj:
                            value += '^^<' + obj['datatype'] + '>'
                        elif 'xml:lang' in obj:
                            value += '@' + obj['xml:lang']
                        row[col_map[var]] = value
                    table.append(row)
            else:
                logging.warning("Return code %s for query '%s'",
                                resp.status,
                                query)
                logging.warning("Message: %s...",
                        re.sub(b"SPARQL-QUERY:[\s\S]*?java\.util",
                               b"SPARGEL-PUREE: ... java.util",
                               resp.data)[:100])
                table = None
                times.query_time = min(timeout, times.client_time)
        except MaxRetryError as e:
            logging.warning('Query timed out')
            times.client_time = (time.time() - start)
            times.query_time = timeout
            table = None
        except ValueError as e:
            logging.warning("Error executing query: %s.", query)
            logging.warning("Headers: %s.", resp.headers)
            logging.warning("Error: %r.", e)
            times.client_time = (time.time() - start)
            times.query_time = timeout
            # HACK(Hannah 04.05.2020): avoid crash
            table = None
            # raise BackendError()
        except:
            logging.warning("Error executing query: %s.", query)
            logging.warning("Headers: %s.", resp.headers)
            times.client_time = (time.time() - start)
            times.query_time = timeout
            # HACK(Hannah 04.05.2020): avoid crash
            table = None
            # raise BackendError()

        logging.debug("Processed Result %r", table)
        return table, {}, times


def main():
    """
    Tiny sanity check when run directly
    """
    sparql = Blazegraph('panarea.informatik.privat', '9999',
            '/bigdata/sparql')
    query = '''
        PREFIX schema: <http://schema.org/>
        PREFIX wikibase: <http://wikiba.se/ontology#>
        PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
        PREFIX wd: <http://www.wikidata.org/entity/>
        PREFIX p: <http://www.wikidata.org/prop/>
        PREFIX wdt: <http://www.wikidata.org/prop/direct/>

        SELECT ?qlc_predicate (COUNT(?city) AS ?qlc_count) WHERE {
         {
          SELECT DISTINCT ?city ?qlc_predicate WHERE {
           ?city ?qlc_predicate ?qlc_free_object .
           ?city wdt:P1082 ?population .
           ?city wdt:P17 wd:Q183 .
           ?city wdt:P31 wd:Q515 .
          }
         }
        }
        GROUP BY ?qlc_predicate
        ORDER BY DESC(?qlc_count)
        LIMIT 128
    '''
    print("trying to query")
    print(json.dumps(sparql.query(query), indent=2))


if __name__ == '__main__':
    main()
