#!/usr/bin/env python3
"""
 Implements a dummy SPARQL backend that always returns empty results.
 Useful for debugging the autocompletion scripts.

 Copyright (c) 2020 University of Freiburg
 Chair of Algorithms and Data Structures
 Author: Johannes Kalmbach <johannes.kalmbach@gmail.com>

"""

from typing import List, Optional, Dict, Any, Set, DefaultDict, NamedTuple, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
import json

from . import Backend, BackendError

class DummyBackend(Backend):
    """
    Qlever abstracts the QLever SPARQL endpoint
    """
    def __init__(self, backend_host,
                 backend_port,
                 backend_url='/',
                 clear_cache=False):
        super().__init__()
        self.is_qlever = False
        self.clear_cache = clear_cache

        # those are required to be set after a call o query()
        self.last_query_compute_time = 5.0  # type: Optional[float]
        self.last_query_client_time = 5.0


    def clear_server_cache(self, clearPinned=False) -> bool:
        """
        Instructs the backend server to clear its cache
        """
        return True


    def query(self, query: str, **kwargs) \
            -> T[Optional[List[List[str]]], Dict[str, Any]]:

        """Returns the result table of the query as a list of rows.

        :param query:
        :return:
        """
        return None, {"error" : "Dummy backend never returns results"}

def main():
    """
    Tiny sanity check when run directly
    """
    sparql = DummyBackend('galera.informatik.privat', '7001', '/')
    query = '''
    PREFIX wd: <http://www.wikidata.org/entity/>
    PREFIX wdt: <http://www.wikidata.org/prop/direct/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT DISTINCT ?dateOfBirth ?astronautLabel WHERE {
      ?astronaut wdt:P106 wd:Q11631 .
      ?astronaut wdt:P27 wd:Q183 .
      ?astronaut wdt:P569 ?dateOfBirth .
      ?astronaut rdfs:label ?astronautLabel .
      FILTER langMatches(lang(?astronautLabel), "en") .
    }
    ORDER BY ASC(?dateOfBirth)
    '''
    print(json.dumps(sparql.query(query)))


if __name__ == '__main__':
    main()
