#!/usr/bin/env python3
"""
 Implements a SPARQL backend using the QLever API

 Copyright (c) 2018 University of Freiburg
 Chair of Algorithms and Data Structures
 Author: Niklas Schnelle

"""

from typing import List, Optional, Dict, Any, Set, DefaultDict, NamedTuple, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
import json
import math
import time
import re
import logging
from urllib3 import HTTPSConnectionPool, HTTPConnectionPool, Retry, make_headers
from urllib3.util import Timeout
from urllib3.exceptions import MaxRetryError
import certifi

from . import Backend, BackendError, QueryTime


NUM_UNIT_REGEX = re.compile(r'(\d+\.?\d*)\s*(\w+)?')

def filter_results_language(results, language):
    """Remove results that contain a literal with another language.

    Empty language is allowed!
    :param results:
    :param language:
    :return:
    """
    langsuffix = "@"+language
    filtered_results = []
    for row in results:
        contains_literal = False
        for value in row:
            value = value.strip()
            if value.startswith('"'):
                contains_literal = True
                if value.endswith(langsuffix):
                    filtered_results.append(row)
        if not contains_literal:
            filtered_results.append(row)
    return filtered_results


class Qlever(Backend):
    """
    Qlever abstracts the QLever SPARQL endpoint
    """
    def __init__(self, backend_host,
                 backend_port,
                 backend_url='/',
                 default_timeout_seconds = 20,
                 clear_cache=False, ssl=False):
        super().__init__()
        self.is_qlever = True
        self.backend_host = backend_host
        self.backend_port = backend_port
        self.backend_url = backend_url
        self.clear_cache = clear_cache
        self.ssl = ssl
        self.default_timeout = default_timeout_seconds
        self.connection_pool = None  # type: HTTPSConnectionPool
        self._init_connection_pool(4)

        self.last_query_compute_time = None  # type: Optional[float]

    def _init_connection_pool(self, pool_maxsize):
        # hardcoded timeout of 120 currently
        self.connection_pool = self.create_connection_pool(pool_maxsize, self.default_timeout)

    def create_connection_pool(self, pool_maxsize, timeout):
        # Retry on 404 and 503 messages because
        # these seem to happen sometimes, but very rarely.
        retry = Retry(total=0, status_forcelist=[404, 503],
                      backoff_factor=0.1)
        timeout *= 1.0
        timeout = Timeout(connect=timeout, read=timeout)
        if self.ssl:
            connection_pool = HTTPSConnectionPool(self.backend_host,
                                                  port=self.backend_port,
                                                  maxsize=pool_maxsize,
                                                  timeout=timeout,
                                                  retries=retry,
                                                  cert_reqs='CERT_REQUIRED',
                                                  ca_certs=certifi.where())
        else:
            connection_pool = HTTPConnectionPool(self.backend_host,
                                                 port=self.backend_port,
                                                 maxsize=pool_maxsize,
                                                 timeout=timeout,
                                                 retries=retry)
        return connection_pool

    def clear_server_cache(self, clearPinned=False) -> bool:
        """
        Instructs the backend server to clear its cache
        """
        params = {
            "cmd": "clearcachecomplete" if clearPinned else "clearcache"
        }

        pool = self.create_connection_pool(2, 200)

        resp = pool.request('GET',
                                            self.backend_url,
                                            headers=make_headers(
                                            keep_alive=False),
                                            fields=params)

        try:
            if resp.status == 200:
                return True
            return False
        except ValueError:
            logging.warning("Error executing cache clearing request")
            logging.warning("Headers: %s.", resp.headers)
            logging.warning("Data: %s.", resp.data)
            return False

    @staticmethod
    def parse_server_timespans(spanstr: str) -> float:
        """
        Parse the timespans in server answers e.g. '128.3ms' to float seconds

        >>> print(round(Qlever.parse_server_timespans('128.3ms'), 4))
        0.1283
        """
        match = NUM_UNIT_REGEX.match(spanstr)
        if not match:
            raise ValueError('Did not match NUM_UNIT_REGEX')
        numstr, unit = match.groups()
        num = float(numstr)

        if not unit:
            return num

        if unit == 'ms':
            return num / 1000
        if unit in ('us', 'μs'):
            return num / 1000000
        if unit == 'ns':
            return num / 1000000000
        raise ValueError('Unsupported timespan unit: '+unit)

    def query(self, query: str, **kwargs) \
            -> T[Optional[List[List[str]]], Dict[str, Any], QueryTime]:
        """Returns the result table of the query as a list of rows.

        :param query:
        :return:
        """
        params = {
            "query": query,
            "send": 100
        }

        times = QueryTime()

        silent = kwargs.get("silent", False)


        timeout = kwargs.get('timeout', self.default_timeout)
        #TODO  : give qlever half a second extra to see where the timeout is violated
        #TODO: this is now really too much, but only for debugging
        extra_time = 0
        http_timeout = timeout + extra_time

        # give qlever some more time to internally calculate P31 and the like
        #if not "timeout" in kwargs:
        #    timeout += 2

        pool = self.connection_pool if timeout is None else self.create_connection_pool(2, http_timeout)
        if timeout is not None:
            #params["timeout"]=math.ceil(timeout)
            params["timeout"]=max(timeout, 30.0) #qlever should now support partial seconds

        if self.clear_cache:
            params["cmd"] = "clearcache"

        if kwargs.get('pin_result', False):
            params['pinresult'] = 'true'

        start = time.time()
        table = []  # type: Optional[List[List[str]]]
        runtime_info = {}  # type: Dict[str, Any]

        try:
            resp = pool.request('GET',
                                                self.backend_url,
                                                headers=make_headers(
                                                    keep_alive=False),
                                                fields=params)
            times.client_time = (time.time() - start)

            if resp.status == 200:
                data = json.loads(resp.data.decode('utf-8'))
                if 'res' not in data:
                    if not silent:
                        print('No query results in backend answer')
                        print(data)
                    table = None
                    """
                    times.compute_time = None
                    times.client_time = None
                    times.query_time = None
                    return None, runtime_info, times
                    """
                else:
                    table = data['res']
                if 'runtimeInformation' in data:
                    runtime_info = data['runtimeInformation']
                    if not runtime_info :
                        runtime_info = ("noRuntime", data)

                else :
                    runtime_info = ("noRuntime", data)

                total_query_time = data['time']['total']
                compute_query_time = data['time']['computeResult']

                times.query_time = Qlever.parse_server_timespans(
                    total_query_time)

                times.compute_time = \
                    Qlever.parse_server_timespans(compute_query_time)
            else:
                logging.warning("Return code %s for query '%s'",
                                resp.status,
                                query)
                logging.warning("Message: %s", resp.data)
                table = None
                times.compute_time = None
        except MaxRetryError as e:
            logging.warning('Query timed out')
            #logging.warning('timing was ', str(timeout))
            #logging.warning('Qlevers timeout not triggered for query:')
            #logging.warning(query)
            times.client_time = (time.time() - start)
            times.query_time = http_timeout
            return None, runtime_info, times
        except ValueError:
            logging.warning("Error executing query: %s.", query)
            logging.warning("Headers: %s.", resp.headers)
            logging.warning("Data: %s.", resp.data)
            times.client_time = (time.time() - start)
            times.query_time = timeout
            return None, runtime_info, times
        except:
            logging.warning("Error executing query: %s.", query)
            logging.warning("Headers: %s.", resp.headers)
            logging.warning("Data: %s.", resp.data)
            times.client_time = (time.time() - start)
            times.query_time = timeout
            return None, runtime_info, times

        logging.debug("Processed Result %r", table)
        return table, runtime_info, times


def main():
    """
    Tiny sanity check when run directly
    """
    sparql = Qlever('galera.informatik.privat', '7001', '/')
    query = '''
    PREFIX wd: <http://www.wikidata.org/entity/>
    PREFIX wdt: <http://www.wikidata.org/prop/direct/>
    PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
    SELECT DISTINCT ?dateOfBirth ?astronautLabel WHERE {
      ?astronaut wdt:P106 wd:Q11631 .
      ?astronaut wdt:P27 wd:Q183 .
      ?astronaut wdt:P569 ?dateOfBirth .
      ?astronaut rdfs:label ?astronautLabel .
      FILTER langMatches(lang(?astronautLabel), "en") .
    }
    ORDER BY ASC(?dateOfBirth)
    '''
    print(json.dumps(sparql.query(query)))


if __name__ == '__main__':
    main()
