#!/usr/bin/env python3
"""
Easy handling of evaluation data such as timings and RuntimeInformation
"""
from typing import TextIO, List, Union, Optional, Dict, Any   # noqa: F401, E501 pylint: disable=unused-import
from copy import deepcopy

import yaml
import yaml.representer
import numpy as np

from query_model import SelectQuery
import query_model


class NamedQuery(SelectQuery):  # pylint: disable=too-few-public-methods
    """
    A SelectQuery extended with an id
    """
    def __init__(self, orig: SelectQuery):
        super().__init__(**orig.__dict__)
        self.name = 'missing-name'  # type: str


def query_representer(dumper, data: SelectQuery):
    """
    Represent SelectQuery types with nice folding
    """
    return dumper.represent_scalar(u'tag:yaml.org,2002:str',
                                   data.to_sparql(), style='|')


yaml.add_representer(SelectQuery, query_representer)


yaml.add_representer(NamedQuery, query_representer)


yaml.add_representer(dict,
                     lambda self, data:
                     yaml.representer.SafeRepresenter.
                     represent_dict(self, data.items()))


class QueryRecord:
    """
    Records evaluation data for a single SPARQL query such as the timing and
    RuntimeInformation (generated by QLever)
    """

    def __init__(self):
        self.query_name = None  # type: str
        self.query_time = None  # type: Optional[float]
        self.query_client_time = None  # type: Optional[float]
        self.query = None  # type: Optional[Union[SelectQuery,str]]
        self.runtime_info = {}  # type: Dict[str, Any]

    @staticmethod
    def header() -> str:
        """
        Returns the header description
        """
        return '\t'.join(
            [
                'Query ID',
                'Query Time (ms)',
                'Total Client Time (ms)'
            ]) + '\n'

    def tsv(self) -> str:
        """
        Returns self in TSV format
        """
        return '\t'.join(
            [
                value_as_str(self.query_name),
                value_as_str(self.query_time*1000
                             if self.query_time else None),
                value_as_str(self.query_client_time*1000
                             if self.query_client_time else None)
            ])+'\n'


class CompletionQueryRecord(QueryRecord):
    """
    Records the evaluation data for a single
    completion query
    """
    def __init__(self):
        super().__init__()
        self.completion_query_name = None  # type: str
        self.pattern = []  # type: List[Any]
        self.orig_pattern = []  # type: List[Any]
        self.alias = None  # type: str
        self.prefix = None  # type: str
        self.pattern_idx = None  # type: int
        self.num_suggestions = None  # type: Optional[int]
        self.num_completion_moves = None  # type: Optional[int]
        self.prefix_length = None # type: Optional[int]
        self.is_warmup = False
        self.has_context = False
        self.is_new_token = False
        self.agnostic_was_used = False
        self.dummy_query = False
        self.is_last_query_of_token = False
        self.certify_target_failed = False
        self.certify_page_failed = False

    @staticmethod
    def header() -> str:
        """
        Returns the header description
        """
        return '\t'.join(
            [
                'Query ID',
                'Completion ID',
                'Triple',
                'Partial Triple',
                'Alias',
                'Triple Num',
                'Has Context',
                'Query Time (ms)',
                '# Suggestions',
                'Position in Result',
                'Prefix Length',
                'Total Client Time (ms)',
                'Is Warmup',
                'First of Token',
                'Query Type',
                'Variable Query',
                'Has Failed',
                'Last of Token',
                'Certify Target Failed',
                'Certify Page Failed'
            ]) + '\n'

    def deduplicate(self):
        def list_to_single(cls, idx):
            d = cls.__dict__
            relevant_keys = ["runtime_info", "query", "query_time", "query_client_time", "num_completion_moves", "has_failed", "certify_target_failed", "certify_page_failed"]
            for el in relevant_keys:
                if isinstance(d[el], list) or isinstance(d[el], tuple):
                    if (len(d[el]) < idx + 1):
                        raise RuntimeError
                    d[el] = d[el][idx]
            return cls
        a = deepcopy(self)
        b = deepcopy(self)
        if len(self.num_completion_moves) == 1:
            res = [list_to_single(a, 0)]
            res[0].type = "single"
            res[0].completion_query_name += "-" + res[0].type
            return res
        else:
            res= [list_to_single(a, 0), list_to_single(b, 1)]
            res[0].type = "sensitive"
            res[1].type = "agnostic"
            res[0].is_last_query_of_token = False
            res[1].is_new_token = False
            res[0].completion_query_name += "-" + res[0].type
            res[1].completion_query_name += "-" + res[1].type
            return res
    def tsv(self) -> str:
        """
        Returns self in TSV format
        """
        #TODO: this is not the right place to do this, but it works.
        if self.alias=="":
            self.dummy_query=True
        return '\t'.join(
            [
                value_as_str(self.query_name),
                value_as_str(self.completion_query_name),
                value_as_str(self.orig_pattern),
                value_as_str([
                    atom if atom != '□'
                    else value_as_str(self.prefix)+'□'
                    for atom in self.pattern]),
                value_as_str(self.alias),
                value_as_str(self.pattern_idx),
                value_as_str(int(self.has_context)),
                value_as_str(self.query_time*1000
                             if self.query_time else None),
                value_as_str(self.num_suggestions),
                value_as_str(self.num_completion_moves),
                value_as_str(self.prefix_length),
                value_as_str(self.query_client_time*1000
                             if self.query_client_time else None),
                value_as_str(int(self.is_warmup)),
                value_as_str(int(self.is_new_token)),
                value_as_str(self.type),
                value_as_str(int(self.dummy_query)),
                value_as_str(int(self.has_failed)),
                value_as_str(int(self.is_last_query_of_token)),
                value_as_str(int(self.certify_target_failed)),
                value_as_str(int(self.certify_page_failed)),

            ])+'\n'


class EvalData:
    """
    Data for evaluation
    """

    def __init__(self, recordClass, eval_log: TextIO, query_log: TextIO):
        self._record_class = recordClass
        self.num_completion_queries = 0
        self.num_completion_moves = 0
        self.num_completion_atoms = 0
        self.num_completion_steps = 0
        self.num_manual_moves = 0
        self.num_failed = 0
        self.num_skipped = 0
        self.query_records = []  # type: List[QueryRecord]
        self.current_record = self._record_class()
        self.eval_log = eval_log
        self.query_log = query_log
        self.steps_until_completion_map = dict()
        self.stuck_targets = []
        self.needed_completion = 0

    def write_log_header(self):
        """
        Writes the log header
        """
        self.eval_log.write(self._record_class.header())

    def finish_log_record(self):
        """
        Writes the current query_record and create a new one initialized with
        its data. The latter allows only updating per query information while
        keeping longer lived data such as the target query
        """
        record = deepcopy(self.current_record)
        # the second, third.... query per Token does not have the new_token bit
        self.current_record.is_new_token=False
        self.current_record.is_last_query_of_token = False
        split_records = record.deduplicate()
        self.query_records.extend(split_records)
        for x in split_records:
            self.eval_log.write(x.tsv())
            self.eval_log.flush()
            query_name = x.completion_query_name \
                if self._record_class == CompletionQueryRecord \
                else x.query_name

            if isinstance(x.runtime_info, tuple):
                x.runtime_info = list(x.runtime_info)

            yaml.dump(
                {
                    "query": query_name,
                    "sparql": x.query,
                    "runtime_info": x.runtime_info
                },
                stream=self.query_log,
                explicit_start=True,
                explicit_end=True)

    def _print_completion_stats(self):
        """
        Prints completion specific statistics if this EvalData records
        a completion session and nothing otherwise
        """
        if self._record_class == CompletionQueryRecord:
            average_completion_moves = self.num_completion_moves / \
                self.num_completion_atoms
            average_manual_moves = self.num_manual_moves / \
                self.num_completion_atoms
            print('Average completion moves: {:.2f} (vs {:.2f} manually)'.
                  format(average_completion_moves, average_manual_moves))

    def print(self):
        """
        Print the most important evaluation data in a human friendly fashion
        """
        query_times = np.array([qr.query_time * 1000 if qr.query_time else 0
                                for qr in self.query_records])
        total_query_time = np.sum(query_times)

        self._print_completion_stats()

        print('Executed {} queries in {:.3f} seconds'.format(
            len(query_times), total_query_time))
        print('{} failed and {} were skipped.'.format(
            self.num_failed, self.num_skipped))
        print('with the following timings')
        print()
        num_sucessful_queries = len(query_times) - self.num_failed
        if num_sucessful_queries > 0:
            print('Average: {:.2f} ms/query'.format(
                total_query_time/(num_sucessful_queries)))
            print('Max: {:.2f} ms/query'.format(np.max(query_times)))
            percentiles = [90, 70, 50]
            percentile_values = np.percentile(query_times,
                                              percentiles,
                                              interpolation='higher')
            for i, percentile in enumerate(percentiles):
                print('{}th percentile: {:.2f} ms/query'.format(
                    percentile, percentile_values[i]))
        else:
            print('No timings available all queries failed')

        print("Number of atoms that needed completion : {}".format(self.needed_completion))
        print("Distribution of typed letters until completion {}".format(self.steps_until_completion_map))
        print("List of stuck targets: {}".format(self.stuck_targets))
        print("Number of stuck targets: {}".format(len(self.stuck_targets)))
        print("Number of successfully completed atoms : {}".format(sum( x for x in self.steps_until_completion_map.values())))

def value_as_str(value: Optional[Union[str, float, list]]) -> str:
    """
    Returns a string representation for the value to be used
    in TSV output

    >>> value_as_str(None)
    ''
    >>> value_as_str([])
    ''
    >>> value_as_str(1.374)
    '1.37'
    >>> value_as_str('foo')
    'foo'
    >>> value_as_str('foo'+chr(9)+'bar')  # No \t in doctest turns into spaces
    'foo   bar'
    >>> value_as_str(['foo', 1.421, None])
    'foo 1.42 '
    """
    if value is None:
        return ''

    if isinstance(value, str):
        return value.replace('\t', '   ')
    if isinstance(value, query_model.query_model.PropertyPath):
        return value.to_sparql().replace('\t', '   ')

    if isinstance(value, float):
        return '{0:.2f}'.format(value)

    if isinstance(value, int):
        return '{}'.format(value)

    if isinstance(value, list):
        return ' '.join(
            [value_as_str(val)
             for val in value])

    return value
