#!/usr/bin/env python3
"""
Pretty printer for Completion Query Logs
"""

import os
import sys
import argparse
from typing import Dict, Any

import yaml


def banner(string: str, width=80) -> str:
    """
    Returns the string padded like ==- string -== to a width of width
    """
    return ('- ' + string + ' -').center(width, '=')


def pretty_runtime_info(rti: Dict[str, Any], indent: int = 1) -> str:
    """
    Prettify the QLever runtime information
    """
    def indenter(num: int, string: str = '') -> str:
        return '│  ' * num + string + '\n'
    res = indenter(indent)
    res += indenter(indent - 1, '├─ ' + rti['description'])
    res += indenter(indent, 'result_size: {:,} × {:,}'.format(
        rti['result_rows'], rti['result_cols']))
    res += indenter(indent, 'columns: {}'.format(
        ', '.join(rti['column_names'])))
    res += indenter(indent, 'total_time: {} ms'.format(
        int(rti['total_time'])))
    res += indenter(indent, 'operation_time: {} ms'.format(
        int(rti['operation_time'])))
    res += indenter(indent, 'cached: {}'.format(rti['was_cached']))

    if 'details' in rti and rti['details']:
        for detail_name, detail_value in rti['details'].items():
            res += indenter(indent, '  {}: {}'.format(
                detail_name, detail_value))

    if rti['children']:
        res += indenter(indent, '┬')
        for child_rti in rti['children']:
            res += pretty_runtime_info(child_rti, indent + 1)

    return res


def main():
    """
    Do the pretty printing
    """
    parser = argparse.ArgumentParser(
        description='Pretty print QLever\'s completion queries')
    parser.add_argument('--input-path', type=str, default='output')
    parser.add_argument('--knowledge-base', default='wikidata-full',
                        help='The knowledge base used')

    args = parser.parse_args(sys.argv[1:])
    path = args.input_path
    knowledge_base = args.knowledge_base
    with open(os.path.join(path, knowledge_base + '.completion-queries.yml'),
              'rt', encoding='utf8') as yml_file:
        comp_queries_yml = yaml.load_all(yml_file)
        for query in comp_queries_yml:
            print(banner(query['query']))
            print('SPARQL:\n\n' + query['sparql']+'\n')
            print('Runtime Information:\n\n' +
                  pretty_runtime_info(query['runtime_info']))


if __name__ == '__main__':
    main()
