#!/usr/bin/env python3
"""
Tree model of a SPARQL query designed for manipulating and printing it
"""
from typing import List, Optional, Dict, Union, Set, DefaultDict, Any, Tuple as T  # noqa: F401, E501 pylint: disable=unused-import
from collections import defaultdict
from copy import deepcopy
from bisect import bisect_left
from abc import ABC, abstractmethod
import query_model
import re


class SPARQLElement(ABC):
    """
    Abstract base class for all elements in the model
    """

    @abstractmethod
    def to_sparql(self, level: int = 0):
        """
        Retunrs the SPARQL for this element
        """


class Prefix(SPARQLElement):
    """
    PREFIX
    """

    def __init__(self, name: str, iri: str):
        self.name = name[:-1]  # token is is "name:" so remove ":"
        self.iri = iri

    def __lt__(self, other):
        if isinstance(other, str):
            return self.iri < other
        return self.iri < other.iri

    def __hash__(self):
        return (self.name + self.iri).__hash__()

    def __eq__(self, other):
        return self.name == other.name and \
                self.iri == other.iri

    def to_sparql(self, level: int = 0) -> str:
        """
        Output Prefix and its children as SPARQL
        """
        return 'PREFIX ' + str(self.name) + ': ' + str(self.iri) + '\n'


class Prologue(SPARQLElement):
    """
    PROGLOGUE
    """
    def __init__(self, base: str, prefixes: List[Prefix]):
        self.base = base
        self.prefixes = []
        self.prefixes_by_name = {}
        for p in prefixes:
            if p.name in self.prefixes_by_name:
                if p != self.prefixes_by_name[p.name]:
                    raise ValueError(
                       "Redefinition of Sparql Prefix with different value"
                       + " is not allowed"
                       + " (" + p.name + ": "
                       + self.prefixes_by_name[p.name].iri
                       + " -> " + p.iri + ")")
            else:
                self.prefixes_by_name[p.name] = p
                self.prefixes.append(p)
        self.prefixes.sort()

        # remove duplicate prefixes


    def merge(self, other: 'Prologue'):
        """
        Merge this prologue with other ignoring base
        """
        added_prefixes = []
        # Remove those that collide with our prefixes
        for other_prefix in other.prefixes:
            if other_prefix.name not in self.prefixes_by_name:
                copied = deepcopy(other_prefix)
                added_prefixes.append(copied)
                self.prefixes_by_name[copied.name] = copied
        self.prefixes = sorted(self.prefixes + added_prefixes)

    def full_iri(self, short: str) -> str:
        """
        For a short name like wd:Q42 gets the full iri <http://wiki…> if short
        is not shorted (i.e. does not include a ':') or the prefix is unknown
        returns it unchanged
        """

        if short.startswith('"') or short.startswith("'"):
            # handle literals that have found their way here
            return short
        short = str(short)
        if ':' not in short:
            return short
        prefix, suffix = short.split(':')
        if prefix not in self.prefixes_by_name:
            return short
        iri = self.prefixes_by_name[prefix].iri
        suffix = re.sub(r'\\(.)', r'\1', suffix)
        return iri[:-1] + suffix + iri[-1]

    def shortname(self, iri: str) -> str:
        """
        Given a fully qualified iri <prefix/suffix> returns the shortened
        prefix:suffix version if there is a fitting prefix otherwise returns
        the iri unchanged
        """
        result = iri
        pos = bisect_left(self.prefixes, iri[:-1])
        if pos == 0:
            return result
        # we want the rightmost element ess than iri
        prefix = self.prefixes[pos - 1]
        prefix_iri_no_gt = prefix.iri[:-1]
        if iri.startswith(prefix_iri_no_gt):
            result = prefix.name + ':' + iri[len(prefix_iri_no_gt):-1]
        return result

    def __eq__(self, other):
        return self.base == other.base and \
                self.prefixes == other.prefixes

    def to_sparql(self, level: int = 0) -> str:
        """
        Output Prologue and its children as SPARQL
        """
        res = ''
        if self.base:
            res += self.base
        if self.prefixes:
            for prefix in self.prefixes:
                res += prefix.to_sparql()
        return res


class GraphPattern(SPARQLElement):
    """
    An abstract graph pattern
    """
    def __init__(self):
        self.parent = None  # type: Optional[GraphPattern]

    @abstractmethod
    def patterns(self) -> 'GroupGraphPattern':
        """
        Returns a list of all subpatterns of this pattern
        """

    @abstractmethod
    def replace(self, orig, rep: str):
        """
        Replaces all occurences of orig in all subpatterns
        with rep
        """

    @abstractmethod
    def all_variables(self) -> Set[str]:
        """
        get all variables that are somehow used in this pattern
        :return:
        """

    @staticmethod
    def pattern_to_sparql(pattern: 'GraphPattern', level: int = 0) -> str:
        """
        Output a pattern as SPARQL
        """
        res = ''
        if isinstance(pattern, list):
            valid = True
            to_print = []
            for element in pattern:
                if isinstance(element, str):
                    if element == '□':
                        valid = False
                    to_print.append(element)
                elif isinstance(element, list):
                    pattern = GroupGraphPattern(element)
                    to_print.append(
                        GraphPattern.pattern_to_sparql(pattern, level + 1))
                elif isinstance(element, PropertyPath):
                    to_print.append(element.to_sparql())
                else:
                    to_print.append('')
                    valid = False
            res = ' '*level + ' '.join(to_print)
            if valid:
                res += ' .\n'
            else:
                res += '\n'
        elif isinstance(pattern, str):
            res = ' '*level + pattern + ' .\n'
        elif hasattr(pattern, 'to_sparql'):
            res = pattern.to_sparql(level)
        else:
            res = 'type: '+ repr(type(pattern)) + ' = ' +repr(pattern)
        return res

    @staticmethod
    def triples(pattern: 'GraphPattern') -> Set[T[str, str, str]]:
        """
        Recursively retrieves all triples in this pattern and its subpatterns
        """
        res = set()  # type: Set[T[str, str, str]]
        for subpattern in pattern.patterns():
            if isinstance(subpattern, list):
                res.add(tuple(subpattern))
            elif isinstance(subpattern, str):
                continue
            else:
                res = res.union(GraphPattern.triples(subpattern))
        return res


    def shortname(self, iri: str) -> str:
        """
        For a short name like wd:Q42 gets the full iri <http://wiki…>
        """
        return self.prologue().shortname(iri)

    def full_iri(self, shortname: str) -> str:
        """
        Given a fully qualified iri <prefix/suffix> returns the shortened
        prefix:suffix version if there is a fitting prefix otherwise returns
        the iri unchanged
        """
        return self.prologue().full_iri(shortname)

    def prologue(self) -> Prologue:
        """
        Returns the Prologue relevant for this GraphPattern
        """
        # Only SelectQueries don't have a parent after being build and they
        # override
        assert self.parent is not None
        return self.parent.prologue()

    def as_graph(self) \
            -> DefaultDict[T[str, str, str], List[T[str, T[str, str, str]]]]:
        """
        Return the triple graph corresponding to this GraphPattern
        where each node is a triple (in tuple form to be hashable)
        and each edge is a variable shared between two triples
        """
        var_to_pattern = defaultdict(list) \
            # type: DefaultDict[str, List[T[str, str, str]]]
        for pattern in self.patterns():  # type: T[str, str, str]
            if not isinstance(pattern, list) or isinstance(pattern, GroupGraphPattern):
                continue

            for atom in pattern:
                if isinstance(atom, query_model.query_model.PropertyPath):
                    all_vars = atom.all_variables()
                    for v in all_vars:
                        var_to_pattern[v].append(tuple(pattern))
                    continue
                elif not atom.startswith('?'):
                    continue
                var_to_pattern[atom].append(tuple(pattern))
        graph = defaultdict(list) \
            # type: DefaultDict[T[str,str,str], List[T[str, T[str,str,str]]]]
        for var, triples in var_to_pattern.items():
            for triple in triples:
                graph[triple].extend(
                    [(var, neighbor) for neighbor in triples
                     if neighbor != triple])
        return graph


class GroupClause(SPARQLElement):
    """
    GROUP including HAVING
    """
    def __init__(self, selectors: List[str], having: Optional[str]):
        self.selectors = selectors
        self.having = having

    def __eq__(self, other):
        return self.selectors == other.selectors and \
                self.having == other.having

    def replace(self, orig, rep: str):
        """
        Replaces the substring orig by rep in all parts of the group clause
        """
        self.selectors = \
            [selector.replace(orig, rep) for selector in self.selectors]
        if self.having:
            self.having.replace(orig, rep)

    def to_sparql(self, level: int = 0) -> str:
        """
        Output GroupClause and its HAVING as SPARQL
        """
        res = ' '*level + 'GROUP BY ' + ' '.join(str(s) for s in self.selectors) + '\n'
        if self.having:
            res += ' '*level + 'HAVING '+self.having + '\n'
        return res


class GroupGraphPattern(list, GraphPattern):
    """
    GroupGraphPatterns are lists of patterns that are matched as a subgraph

    TODO(schnelle): inheriting from list is a bit risky as we could miss
    a pattern being added and thus setting us as the parent
    """
    def __init__(self,
                 iterable: List[Union[GraphPattern, List[str], str]] = None):

        self.completion_triple_is_reversed = False
        if iterable:
            list.__init__(self, iterable)
        GraphPattern.__init__(self)
        for pattern in self:
            if isinstance(pattern, GraphPattern):
                pattern.parent = self

    def extend(self, iterable: Any):
        for pattern in iterable:
            self.append(pattern)

    def append(self, pattern: Union[GraphPattern, list, str]):
        if isinstance(pattern, GraphPattern):
            pattern.parent = self
        super().append(pattern)

    def __setitem__(self, idx, pattern):
        if isinstance(pattern, GraphPattern):
            pattern.parent = self
        super().__setitem__(idx, pattern)

    def to_sparql(self, level: int = 0) -> str:
        res = '{\n'
        if self:
            for pattern in self[:-1]:
                res += self.pattern_to_sparql(pattern, level + 1)
            res += self.pattern_to_sparql(
                self[-1], level + 1)
        res += ' '*level + '}'
        return res

    def patterns(self) -> 'GroupGraphPattern':
        return self

    @staticmethod
    def is_triple(p):
        b1 = isinstance(p, list) and len(p) == 3 and isinstance(p[0], str)
        b2 = b1 and isinstance(p[2], str) and (isinstance(p[1], str) or isinstance(p[1], PropertyPath))
        return b2

    def all_variables(self) -> Set[str]:
        res = set()
        for pattern in self:
            if isinstance(pattern, GraphPattern):
                res.update(pattern.all_variables())
            elif isinstance(pattern, list):
                assert(self.is_triple(pattern))
                if pattern[0].startswith('?'):
                    res.add(pattern[0])
                if pattern[2].startswith('?'):
                    res.add(pattern[2])
                if isinstance(pattern[1], str) and pattern[1].startswith('?'):
                    res.add(pattern[1])
                elif isinstance(pattern[1], PropertyPath):
                    res.update(pattern[1].all_variables())
            elif not isinstance(pattern, Filter) :
                raise ValueError("Children of GroupGraphPatterns must either be GraphPatterns or lists representing triples or Filters")
        return res


    def replace(self, orig, rep: str):
        """
        Replaces the substring orig with rep in all patterns
        """
        for pattern_idx, pattern in enumerate(self):
            if isinstance(pattern, list):
                self[pattern_idx] = \
                    [atom.replace(orig, rep) for atom in pattern]
            else:
                pattern.replace(orig, rep)


class OptionalPattern(GraphPattern):
    """
    OPTIONAL graph pattern
    """

    def __init__(self, pattern: GroupGraphPattern):
        super().__init__()
        self.pattern = pattern
        self.pattern.parent = self

    def __eq__(self, other):
        if not isinstance(other, OptionalPattern):
            return False
        return self.pattern == other.pattern

    def all_variables(self) -> Set[str]:
        return self.pattern.all_variables()

    def replace(self, orig, rep: str):
        """
        Replace all substrings orig in the optional pattern
        with rep
        """
        self.pattern.replace(orig, rep)

    def patterns(self) -> GroupGraphPattern:
        return self.pattern

    def to_sparql(self, level: int = 0) -> str:
        """
        Output this OptionalPattern as SPARQL
        """
        res = ' '*level + 'OPTIONAL '
        res += self.pattern.to_sparql(level+1) + '\n'
        return res

class ValuesPattern(GraphPattern):
    """
    InlineData (aka VALUES clause)
    """
    def __init__(self, variables: List[str], values: List[List[str]]):
        super().__init__()
        self.variables = variables
        self.values = values

    def all_variables(self) -> Set[str]:
        return set(self.variables)

    def __eq__(self, other):
        if not isinstance(other, ValuesPattern):
            return False
        return self.variables == other.variables and self.values == other.values

    def __str__(self):
        return self.to_sparql()

    def replace(self, orig, rep:str):
        for i, el in enumerate(self.variables):
            if el == orig:
                self.variables[i] = rep
        for i, el in enumerate(self.values):
            if el == orig:
                self.values[i] = rep
    def patterns(self) -> 'GroupGraphPattern':
        return GroupGraphPattern()

    def to_sparql(self, level: int = 0):
        res = ' ' * level + "VALUES "
        res += '(' + " ".join(self.variables) + ')'
        res += '{' + " ".join("(" + " ".join(val) + ")" for val in self.values) + '}'
        return res

class BindPattern(GraphPattern):
    """
    Currently bind patterns just consist of their text, they can thus not
    reasonably be autocompleted.
    """
    def __init__(self, expr : str, var: str):
        super().__init__()
        self.expr = expr
        self.var = var
    def __eq__(self, other):
        if not isinstance(other, BindPattern):
            return False
        return self.expr == other.expr and self.var == other.var
    def to_sparql(self, level: int = 0):
        return ' '*level + "BIND(" + self.expr + " as " + self.var + ")"

    def all_variables(self) -> Set[str]:
        # all the variables in the expression have to be
        # in use before for this to make any sense.
        return {self.var}

    def replace(self, orig, rep: str):
        pass

    def patterns(self) -> 'GroupGraphPattern':
        return GroupGraphPattern()

    def __str__(self):
        return self.to_sparql()



class UnionPattern(GraphPattern):
    """
    UNION graph pattern
    """

    def __init__(self, lhs: GroupGraphPattern, rhs: GroupGraphPattern):
        super().__init__()
        self.lhs = lhs
        self.lhs.parent = self
        self.rhs = rhs
        self.rhs.parent = self

    def __eq__(self, other):
        if not isinstance(other, UnionPattern):
            return False
        return self.lhs == other.lhs and self.rhs == other.rhs

    def all_variables(self) -> Set[str]:
        r = self.lhs.all_variables()
        r.update(self.rhs.all_variables())
        return r

    def replace(self, orig, rep: str):
        """
        Replaces the substring orig by rep in both sides of the Union
        """
        self.lhs.replace(orig, rep)
        self.rhs.replace(orig, rep)

    def patterns(self) -> GroupGraphPattern:
        return GroupGraphPattern(self.lhs.patterns() + self.rhs.patterns())

    def to_sparql(self, level: int = 0) -> str:
        """
        Output this union and both its sides as SPARQL
        """
        res = ' '*level + self.lhs.to_sparql(level)
        res += ' UNION '
        res += self.rhs.to_sparql(level) + '\n'
        return res


"""
Helper class, stores the whole selector text
(Sample ?bla As ?blubb)
and the resulting variable (?blubb)
"""
class Selector:
    def __init__(self, text, var):
        self.text = text
        self.var = var

    def __eq__(self, other):
        if not isinstance(other, Selector):
            return False
        return self.text == other.text

    def __hash__(self):
        return self.text.__hash__()

    def __str__(self):
        return self.text
    def __repr__(self):
        return self.text

    def replace(self, orig, rep : str):
        self.text.replace(orig, rep)
        self.var.replace(orig, rep)




class SelectQuery(GraphPattern):
    """
    SELECT query including prologue
    """


    def __init__(self, owned_prologue: Prologue, distinct: bool, selectors: List[Selector],
                 where: GroupGraphPattern, group: Optional[GroupClause],
                 order: Optional[List[str]],
                 limit: Optional[str],
                 parent: Optional['SelectQuery'] = None):
        super().__init__()
        self.parent = parent
        self.distinct = distinct
        self.owned_prologue = owned_prologue
        self.selectors = selectors
        self.where = where
        self.group = group
        self.order = order
        self.limit = limit

        # Set this query as parent of all children, doing
        # this during parsing would require us to create parents
        # before children and only afterwards setting attributes
        self.where.parent = self
        for pattern in self.where:
            if isinstance(pattern, GraphPattern):
                pattern.parent = self.where

    def prologue(self) -> Prologue:
        """
        Return the relevant Prologue, as a SelectQuery we might have one
        ourselves or we might have to ask our parent
        """
        if self.parent:
            return self.parent.prologue()
        return self.owned_prologue

    def all_variables(self) -> Set[str]:
        return self.where.all_variables()

    def __eq__(self, other):
        """
        Compares two SelectQuerys for equality
        """
        if not isinstance(other, SelectQuery):
            return False
        return self.owned_prologue == other.owned_prologue and \
            self.selectors == other.selectors and \
            self.where == other.where and \
            self.group == other.group and \
            self.order == other.order and \
            self.limit == other.limit

    def replace(self, orig, rep: str):
        """
        Replaces the substring orig by rep in all parts of the query
        """
        self.selectors = \
            [selector.replace(orig, rep) for selector in self.selectors]
        self.where.replace(orig, rep)
        if self.group:
            self.group.replace(orig, rep)
        if self.order:
            self.order = \
                [o.replace(orig, rep) for o in self.order]

    def patterns(self) -> GroupGraphPattern:
        return self.where.patterns()

    def to_sparql(self, level: int = 0) -> str:
        """
        Output SelectQuery and its children as SPARQL
        """
        res = ''

        if self.owned_prologue:
            # subqueries don't have a prologue
            res += self.owned_prologue.to_sparql()
            res += '\n'  # for better readability

        if self.parent:
            # if we are a subquery we need to add outer {}
            res += ' '*level + '{\n'
            level += 1

        distinct = 'DISTINCT ' if self.distinct else ''
        res += ' '*level + 'SELECT ' + distinct + ' '.join(
            [str(selector) for selector in self.selectors])
        res += ' WHERE '
        res += self.where.to_sparql(level) + '\n'

        if self.group:
            res += self.group.to_sparql(level)

        if self.order:
            res += ' '*level + 'ORDER BY '+' '.join(self.order)+'\n'
        if self.limit:
            res += ' '*level + 'LIMIT '+str(self.limit)

        if self.parent:
            # if we are a subquery we need to add outer {}
            level -= 1
            res += ' '*level + '}\n'
        return res



class PropertyPath(SPARQLElement):
    class PathElt:
        def __init__(self, pathPrimary, inverse=False, modifier=None):
            self.inverse = inverse
            self.pathPrimary = pathPrimary
            self.modifier = modifier

        def to_sparql(self):
            i = "^" if self.inverse else ""
            mod = self.modifier if self.modifier else ""
            need_braces = False
            return ("(" if need_braces else "") + i + self.pathPrimary + (")" if need_braces else "") + mod

        def reverse(self):
            self.inverse = not self.inverse

        def to_reverse_sparql(self):
            i = "^" if not self.inverse else ""
            mod = self.modifier if self.modifier else ""
            return "(" + i + self.pathPrimary +")" + mod

        def getSingle(self):
            if self.inverse or self.modifier:
                raise ValueError("Trying to get a simple Iri from complex property path " + self.to_sparql())
            return self.pathPrimary


        def remove_modifiers(self):
            return PropertyPath.PathElt(self.pathPrimary, self.inverse, None)

        def all_elements(self):
            return {self.pathPrimary}

        def apply(self, f):
            self.pathPrimary = f(self.pathPrimary)

        def can_be_empty(self):
            return self.modifier == "?" or self.modifier == "*"

        def to_non_empty_path(self):
            """
            Returns a copy where the * modifier is changed to + and ? is omitted,
            forcing the property path to have at least length one
            """
            modifier = self.modifier
            if modifier == "?": modifier = None
            if modifier == "*": modifier = "+"
            return PropertyPath.PathElt(self.pathPrimary, self.inverse, modifier)

        def __str__(self):
            return self.to_sparql()

        def __repr__(self):
            return self.to_sparql()





    class PathSequence():
        def __init__(self, els):
            assert (len(els) > 0)
            self.els = els
        def to_sparql(self):
            return "/".join(x.to_sparql() for x in self.els)
        def reverse(self):
            for x in self.els:
                x.reverse()
            self.els.reverse()
        def to_reverse_sparql(self):
            return "/".join(reversed(x.to_reverse_sparql() for x in self.els))
        def getSingle(self):
            if len(self.els) != 1:
                raise ValueError("Trying to get a simple Iri from complex property path " + self.to_sparql())
            return self.els[0].getSingle()
        def  __len__(self):
            return len(self.els)

        def all_elements(self):
            r = set()
            for el in self.els:
                r.update(el.all_elements())
            return r

        def apply(self, f):
            for el in self.els:
                el.apply(f)

    def __init__(self, alternatives: List[PathSequence]):
        assert (len(alternatives) > 0)
        self.alternatives = alternatives

    def getSingle(self):
        """
        get the single iri, if this path represents a single iri,
        else throw a ValueError
        """
        if len(self.alternatives) != 1:
            raise ValueError("Trying to get a simple Iri from complex property path " + self.to_sparql())
        return self.alternatives[0].getSingle()
    def isSingle(self):
        """
        Does this consist of a single element
        """
        try:
            self.getSingle()
            return True
        except ValueError:
            return False


    def to_sparql(self, level: int = 0):
        return "|".join(x.to_sparql() for x in self.alternatives)

    def reverse(self):
        for x in self.alternatives:
            x.reverse()

    def to_reverse_sparql(self, level: int = 0):
        return "|".join(x.to_reverse_sparql() for x in self.alternatives)

    def all_variables(self):
        # for some reason we need the set + filter stuff here to get the
        # desired result
        return set(list(filter(lambda x: x.startswith("?"), self.all_elements())))

    def apply(self, f):
        for el in self.alternatives:
            el.apply(f)

    def all_elements(self):
        r = set()
        for el in self.alternatives:
            r.update(el.all_elements())
        return r

    def __eq__(self, other):
        if isinstance(other, str):
            return self.to_sparql() == other
        elif isinstance(other, PropertyPath):
            return self.to_sparql() == other.to_sparql()
        else :
            return False
    def __str__(self):
        return self.to_sparql()

    def __hash__(self):
        return self.to_sparql().__hash__()

    def __len__(self):
        return len(self.to_sparql())

    class NotSpeciallError(Exception):
        pass


    def getSpecialPath(self):
        if len(self.alternatives) != 1:
            raise self.NotSpeciallError
        seq = self.alternatives[0]
        if len(seq) != 2 or seq.els[0].inverse or seq.els[0].modifier:
            raise self.NotSpeciallError
        if seq.els[1].inverse or seq.els[1].modifier not in ("*", "?"):
            raise self.NotSpeciallError


        return seq.els[0].pathPrimary, seq.els[1].pathPrimary, seq.els[1].modifier

    """Return a predicate path that returns a superset of the original path
       which is possibly easier to compute """
    def get_superset(self):
        if len(self.alternatives) != 1:
            return self
        seq = self.alternatives[0]
        if len(seq) == 2 :
            return PropertyPath([PropertyPath.PathSequence([seq.els[0].remove_modifiers()]),
                                PropertyPath.PathSequence([seq.els[1].remove_modifiers()])])
        return self

    def remove_modifiers(self):
        if len(self.alternatives) != 1:
            raise ValueError("remove_modifiers should never be called on an alternative for now.")
        seq = self.alternatives[0]
        return PropertyPath([PropertyPath.PathSequence([x.remove_modifiers() for x in seq.els])])



class Filter(SPARQLElement):
    def __init__(self, text, variables):
        self.text = text
        self.variables = variables

    def __str__(self):
        return self.text

    def __repr__(self):
        return self.text

    def to_sparql(self, level: int = 0) -> str:
        return self.text + ' .\n'

    def patterns(self):
        return []

    def __eq__(self, other):
        if not isinstance(other, Filter):
            return False
        return self.text == other.text


class MinusGraphPattern(GraphPattern):
    def __init__(self, pattern : GroupGraphPattern):
        super(MinusGraphPattern, self).__init__()
        self.pattern = pattern
        self.pattern.parent =  self

    def patterns(self):
        return self.pattern
    def replace(self, orig, rep:str):
        self.pattern.replace(orig, rep)
    def all_variables(self):
        return self.pattern.all_variables()
    def __eq__(self, other):
        return isinstance(other, MinusGraphPattern) and self.pattern == other.pattern
    
    def to_sparql(self, level: int = 0):
        return "MINUS " + self.pattern.to_sparql(level + 1)


