#!/usr/bin/env python3
"""
(Simulated) Auto-completion with QLever
"""

from typing import TextIO, List, Union, Optional, Dict, \
        Set, Tuple as T
from collections import namedtuple

from copy import deepcopy
import random
import re

from backend import Backend, BackendError, qlever, query_async_with_timeout
from query_model.query_model import SelectQuery, GroupGraphPattern, \
        GraphPattern, OptionalPattern, UnionPattern, ValuesPattern, PropertyPath, Filter, \
        MinusGraphPattern
import query_model
import concurrent.futures

from helpers import Colors
from eval_data import EvalData, CompletionQueryRecord, NamedQuery
import template_handling
import query_model.qlever_hacks
from itertools import count

TYPING_STEP = 1
INSTANT_VIEWABLE = 7


class CompletionPos:
    def __init__(self, pattern_idx, atom_idx):
        self.pattern_idx = pattern_idx
        self.atom_idx = atom_idx


class SparqlCompleter:
    """
    Simulates auto completion of SPARQL queries
    """

    def __init__(self, comp_templates: Dict[str, str],
                 replace_predicates: Dict[str, str],
                 completion_query_log: TextIO,
                 eval_log: TextIO,
                 show_queries: bool,
                 sparql_backend: Backend, qlever_fallback_backend: Backend,
                 qlever_fallback_templates: Dict[str, str],
                 agnostic_comp_templates: Dict[str, str],
                 agnostic_fallback_timeout: int = None,
                 divArgs = None
                 ):
        self.comp_templates = comp_templates
        self.replace_predicates = replace_predicates
        if self.replace_predicates is None :
            self.replace_predicates = dict()

        #TODO: make this configurable
        self.minimum_key_numbers = 2 # only used for the legacy mode
        self.forced_typing_steps = divArgs.forced_typing_steps
        self.mode = divArgs.mode
        self.show_queries = show_queries
        self.reverse_fixed_object_triples = divArgs.reverse_fixed_object_triples
        self.superset_predicates = divArgs.superset_predicates
        self.exclude_labels = divArgs.exclude_labels
        self.query_variables=divArgs.query_variables
        self.certify_fallback=divArgs.certify_fallback
        self.backend = sparql_backend
        self.qlever_fallback_backed = qlever_fallback_backend
        self.qlever_fallback_templates = qlever_fallback_templates
        if self.backend.is_qlever:
            self.qlever_fallback_backed = self.backend
            self.qlever_fallback_templates = self.comp_templates
        self.agnostic_comp_templates = agnostic_comp_templates
        self.use_agnostic_backup = agnostic_fallback_timeout is not None and divArgs.mode != "sensitive"
        self.mode = divArgs.mode
        self.agnostic_fallback_timeout = agnostic_fallback_timeout
        self.eval = EvalData(
            CompletionQueryRecord,
            eval_log,
            completion_query_log)
        self.completion_query_counter = 0
        # TODO<joka921> read this from the template yamls
        #self.warmup_predicates = {"wdt:P31", "wdt:P31/wdt:P279*", "fbe:is-a", "fb:type.object.type"}
        self.warmup_predicates = {}
        #self.used_predicates = set()  # type: Set[str]

    def _completion_query(self, name: str, use_fallback: bool, only_main_query:bool = False):
        """
        Returns a completion query given its name and the a query
        from which to take the prologue. This is a copy so that the
        query can be manipulated freely
        If no completion query with that name exists None is returned
        returns a list of (completion query, agnostic completion query) where
        the second element is the empty string in case of errors.
        if we have no agnostic templates, the list has only one entry
        """
        if use_fallback:
            if name not in self.qlever_fallback_templates:
                return None
            result = deepcopy(self.qlever_fallback_templates[name])
        else:
            if name not in self.comp_templates:
                return None
            result = deepcopy(self.comp_templates[name])


        agnostic_fallback = "" # empty string allows replacement etc without errors
        if self.use_agnostic_backup and not only_main_query:
            agnostic_fallback = self.agnostic_comp_templates[name]
            return [result, agnostic_fallback]
        else:
            return [result]

    @staticmethod
    def connected_patterns(graph_pattern: GraphPattern,
                           seeds: Set[str], exclude_labels) -> GroupGraphPattern:
        """
        Determine which triples are connected with at least one of the seed
        atoms and also connect to at least one non-variable
        """
        # built reachability graph
        graph = graph_pattern.as_graph()
        # Check which patterns are connected to a seed through variables
        connected = set()  # type: Set[T[str, str, str]]
        # We only care about seeds which are variables
        seeds = {seed for seed in seeds if str(seed).startswith('?')}
        queue = []  # type: List[T[str, str, str]]
        for triple in graph.keys():
            if exclude_labels and triple[2].startswith("?") and isinstance(triple[1], PropertyPath) and triple[1].isSingle() and (triple[1].getSingle() == "rdfs:label" or triple[1].getSingle() == "fb:type.object.name"):
                pass
            elif any([atom in seeds for atom in triple]):
                queue.append(triple)

        while queue:
            current = queue.pop()
            if '□' in current or '' in current:
                continue
            connected.add(current)
            for (_, triple) in graph[current]:
                if triple not in connected:
                    queue.append(triple)
        # Need to annotate the type more loosely because mypy can't match
        # List[List[str]] to List[Union[List[str],…]]
        triples = [list(deepcopy(triple)) for triple in connected] \
            # type: List[Union[GraphPattern, str, List[str]]]
        return GroupGraphPattern(triples)

    def _propagate_context(self,
            completion_pattern: GraphPattern,
            context_pattern: GraphPattern,
            seeds: Set[str],
            replace_predicates: Dict[str, str]):
        """
        Recursively adds patterns from the context to the completion pattern,
        that is triples/UNIONs/OPTIONALs from the user supplied query that form
        the context for the completion.
        The seeds are the variables directly connected to the completion point.
        These are used as the starting points in a DFS on the triple graph to
        determine connected triples withing a GroupGraphPattern
        The context is added at the beginning, since we want the completion
        triple to be in exactly the right context position.
        """

        extension = []

        #remove the variables from the seed
        seeds_remov = {i for i in seeds if not str(i).startswith("?")}

        for s in seeds_remov:
            seeds.remove(s)

        seeds_old = {}

        last_iter = False

        while not last_iter :
            if seeds_old == seeds:
                # we have found no more seeds in the last iteration,
                # we can finalize the result
                last_iter = True
            seeds_old = deepcopy(seeds)
            extension = []


            # we have to keep the order of Triples and optionals since this matters
            # for the semantics and for the optimizability


            # find all the triples that we have to include into the context
            connected_patterns = SparqlCompleter.connected_patterns(
                context_pattern, seeds, self.exclude_labels)

            # also consider all the variables which we are connected to
            for trip in connected_patterns:
                if self.exclude_labels and trip[2].startswith("?") and isinstance(trip[1], PropertyPath) and \
                        trip[1].isSingle() and (
                        trip[1].getSingle() == "rdfs:label" or trip[1].getSingle() == "fb:type.object.name"):
                    continue
                for el in trip:
                    if str(el).startswith("?"):
                        seeds.add(el)


            # Recursively deal with subpatterns
            for pattern in context_pattern.patterns():
                if isinstance(pattern, UnionPattern):
                    lhs_context = GroupGraphPattern()
                    rhs_context = GroupGraphPattern()
                    union = UnionPattern(lhs_context, rhs_context)
                    union.parent = completion_pattern

                    self._propagate_context(
                        lhs_context, pattern.lhs, seeds, replace_predicates)
                    self._propagate_context(
                        rhs_context, pattern.rhs, seeds, replace_predicates)

                    if union.lhs and union.rhs:
                        extension.append(union)
                    continue

                if isinstance(pattern, OptionalPattern):
                    opt_context = GroupGraphPattern()
                    opt = OptionalPattern(opt_context)
                    opt.parent = completion_pattern

                    self._propagate_context(
                        opt_context, pattern.pattern, seeds, replace_predicates)
                    if opt.pattern:
                        extension.append(opt)
                    continue

                if isinstance(pattern, MinusGraphPattern):
                    minus_context = GroupGraphPattern()
                    minus = MinusGraphPattern(minus_context)
                    minus.parent = completion_pattern
                    self._propagate_context(
                        minus_context, pattern.pattern, seeds, replace_predicates
                    )
                    if minus.pattern:
                        extension.append(minus)
                    continue

                if isinstance(pattern, GroupGraphPattern):
                    #TODO<joka921>: this ignores the scoping of GroupGraphPatterns,
                    # but this does not make a difference in the benchmarks.
                    tmp_context = GroupGraphPattern()
                    tmp_context.parent = completion_pattern
                    self._propagate_context(tmp_context, pattern, seeds, replace_predicates)
                    # this is where we ignore the scoping
                    extension.extend(tmp_context)  # GroupGraphPatterns inherit from list.

                if isinstance(pattern, SelectQuery) and any(True for el in pattern.selectors if el.var in seeds):
                    extension.append(pattern)
                    #TODO<which of the variables in the selector do we have to add to
                    # the seeeds, i think all?
                    for var in pattern.selectors:
                        if var.var.startswith("?"):
                            seeds.add(var.var)

                if isinstance(pattern, ValuesPattern) and any(True for el in pattern.variables if el in seeds):
                    extension.append(pattern)

                #TODO: BindPattern, technically we have to check, whether
                # any seed is contained in the evaluated expression,
                # but we currently do not really support them anyway.

                if isinstance(pattern, Filter) and all(el in seeds for el in pattern.variables):
                    extension.append(pattern)


                # a connected triple
                if pattern in connected_patterns :
                    # Replace some predicates with versions more suited for contexts
                    # e.g. rdfs:label with @en@rdfs:label
                    # no replacement for now
                    #predicate = completion_pattern.full_iri(pattern[1])
                    #if predicate in replace_predicates:
                    #    pattern[1] = replace_predicates[predicate]
                    if self.exclude_labels and  pattern[2].startswith("?") and isinstance(pattern[1], PropertyPath) and pattern[1].isSingle() and (pattern[1].getSingle() == "rdfs:label" or pattern[1].getSingle() == "fb:type.object.name"):
                        pass
                        #also include non-connected label triples from the context.
                    else:
                        extension.append(pattern)

            # now we have the complete context in correct order, prepend it
            # TODO<joka921> check for the best way to make this work
            # with immutability/references
        completion_pattern.extend(extension)

    def extend_with_context(self,
            completion_pattern: GraphPattern,
            context_pattern: GraphPattern,
            seeds: Set[str],
            replace_predicates: Dict[str, str]):
        """
        Extends the given completion query (including subqueries) with context
        from the query currently being completed. Incomplete patterns i.e.
        those containing None or completion points are ignored as context
        """
        # For the context of an inner GroupGraphPattern (UNION, OPTIONAL etc)
        # we actually want everything connected inside its grandest parent.
        while context_pattern.parent and not(isinstance(context_pattern, SelectQuery)):
            context_pattern = context_pattern.parent

        # We only really care about those seeds which are actually contained
        # in the completion template e.g. the subject in a predicate template
        # without a fixed object
        #TODO<joka921> eliminate this variable. We always care about all seeds,
        # since the seeds are set manually beforehand.
        for pattern in completion_pattern.patterns():
            if isinstance(pattern, GroupGraphPattern):
                self.extend_with_context(
                    pattern, context_pattern, seeds, replace_predicates)
            elif isinstance(pattern, SelectQuery):
                self.extend_with_context(
                    pattern.where, context_pattern, seeds, replace_predicates)
        self._propagate_context(
            completion_pattern,
            context_pattern,
            seeds,
            replace_predicates)

    @staticmethod
    def has_completion_point(where: GraphPattern) -> bool:
        """
        Checks if there is a completion point in any of the
        triples in the given GraphPattern.
        A completion point is an element marked with '□'
        at which auto completion can be performed
        """
        for pattern in where.patterns():
            if not isinstance(pattern, list):
                # in the future add completing in FILTER etc
                continue
            if '□' in pattern:
                return True
        return False

    def _complete_atom_with_query(self,
                                  current: GroupGraphPattern,
                                  orig_atom: str,
                                  comp_query: SelectQuery,
                                  inverse_property: bool = False,
                                  only_use_agnostic: bool = False,
                                  is_literal=False) \
            -> bool:
        """
        Completes the given atom using the given completion query
        and updates self.num_completion_moves accordingly.
        If the completion doesn't succeed returns a completion point
        marker
        Return true if any of the completion queries (sensitive or agnostic) did not fail
        """
        if isinstance(orig_atom, query_model.query_model.PropertyPath):
            orig_atom = orig_atom.getSingle()

        actual_backend = self.qlever_fallback_backed if self.eval.current_record.is_warmup else self.backend


        # strange hack
        if actual_backend.is_qlever: # only required for the non-agnostic fallback
            comp_query[0] = query_model.qlever_hacks.qlever_property_path_hack(comp_query[0])


        # all the possible ways the the atom might be encoded in qlever and blazegraph.
        # this is actually just relevant because of fbeasy and blazegraph always adding
        # an automated prefix
        if not is_literal:
            orig_atom_full_iri = [current.full_iri(orig_atom)]
            if not actual_backend.is_qlever and actual_backend.is_blazegraph and orig_atom.startswith("fbe:"):
                orig_atom_full_iri.append("<file:/local/data/blazegraph/freebase-easy/" + re.sub(r'\\(.)', r'\1', orig_atom[4:])+ '>')
        else:
            orig_atom_full_iri = [orig_atom]

        #all_triples = GraphPattern.triples(comp_query.where)
       # self.used_predicates = self.used_predicates.union(
       #     [comp_query.full_iri(triple[1]) for triple in all_triples])

            #print("query qlever:")
        # for now, ignore the agnostic stuff until the interface works
        # TODO: apply the stuff here
        if only_use_agnostic or self.mode == "agnostic" or (self.agnostic_fallback_timeout and self.agnostic_fallback_timeout <= 0.0):
            res, rti, times = self.qlever_fallback_backed.query(comp_query[1].to_sparql())
            res, rti, times = [res], [rti], [times]
            comp_query = [comp_query[1]]
        else:
            if self.use_agnostic_backup:
                res, rti, times = query_async_with_timeout(actual_backend, self.qlever_fallback_backed, comp_query[0].to_sparql(), comp_query[1].to_sparql(), self.agnostic_fallback_timeout, comp_query[2].to_sparql() if self.certify_fallback and len(comp_query) == 3 else None, orig_atom_full_iri[0])
            else:
                res, rti, times = actual_backend.query(comp_query[0].to_sparql())
                res, rti, times = [res], [rti], [times]
                comp_query = [comp_query[0]]

        self.eval.current_record.query = [c.to_sparql() for c in comp_query]

        # now res, rti, times are lists which either all have 1 or all have 2 entries

        # The query_client_ime is None if we timed out but we still
        # need to keep track of this
        self.completion_query_counter += 1
        self.eval.num_completion_queries += 1
        self.eval.current_record.query_client_time = \
            [t.client_time for t in times]
        self.eval.current_record.num_suggestions = len(res) if res else 0
        self.eval.current_record.runtime_info = rti
        self.eval.current_record.query_time = \
            [t.client_time for t in times]
        self.eval.current_record.certify_target_failed = [not t.certifyTargetWorked for t in times]
        self.eval.current_record.certify_page_failed = [not t.certifyFirstPageWorked for t in times]



        for idx, t in enumerate(times):
            if t.query_time:
                # Override query_time with server measured time if available
                self.eval.current_record.query_time[idx] = \
                    t.query_time

        # If res is None the query failed and we're done here
        self.eval.current_record.num_completion_moves = [1000000] * len(res)
        self.eval.current_record.has_failed = [True] * len(res)
        for idx, r in enumerate(res):
            if r is None:
                continue
            self.eval.current_record.has_failed[idx] = False
            for res_row_idx, res_row in enumerate(r):
                if self.show_queries:
                    print('{} (count: {})'.format(res_row[0], res_row[1]))
                if res_row_idx == 0 and res_row[0].endswith("continue>"):
                    continue
                if res_row[0] in orig_atom_full_iri and (not inverse_property or res_row[1] == "1"):
                    if self.show_queries:
                        print(Colors.OKGREEN+'Found at {}\n'.format(res_row_idx),
                              Colors.ENDC)
                    #TODO<joka921>:Does this fbeasy-blazegraph hack work for all
                    # knowledgebases

                    self.eval.current_record.num_completion_moves[idx] = 1 + res_row_idx
        return not all(self.eval.current_record.has_failed)

    @staticmethod
    def prefix_regex(qlever_workaround: bool, prefix: str) -> str:
        """
        Creates the prefix regex needed to filter for the given prefix. As an
        example for an IRI prefix <http://foo.ba it creates 'regex(str(?var),
        "^http://foo.bar") while for a literal the str() is not needed and only
        '"' are stripped

        >>> SparqlCompleter.prefix_regex(False, "<http://foo.ba")
        '"^http://foo.ba"'
        >>> SparqlCompleter.prefix_regex(False, '"Ad')
        '"^Ad"'
        """

        # If the prefix contains regex control characters,
        # those have to be escaped with \\ (one for Sparql and one for the
        # regex)

        regex = r"([\[\]^$.|?*+\(\)\\])"
        prefix = re.sub(regex, r'\\\\\1', prefix)
        # QLever currently doesn't implement this correctly
        if qlever_workaround:
            if prefix.startswith('<'):
                return '{}'.format(prefix)
            return '{}'.format(prefix.replace('"', '\\"'))
        # Standard SPARQL
        return '{}'.format(prefix.lstrip('"'))


    def _complete_atom_with_queries(self, current: GroupGraphPattern,
                                    orig: GroupGraphPattern,
                                    orig_atom: str,
                                    comp_query: str,
                                    alias_map: Dict[str, List[str]],
                                    comp_query_something_typed = None) \
            -> Optional[str]:
        """
        Completes the given atom using the given completion query and
        HAVING/FILTER for prefix search to minimize moves
        """
        #TODO<joka921> this disallows completing of property paths

        if self.eval.current_record.is_warmup and not self.backend.is_qlever:
            print("Skipping Token that has no Context for a Virtuoso or Blazegraph evaluation")
            return orig_atom

        only_use_agnostic = False
        never_use_exact_regex = False
        is_literal = False
        if isinstance(orig_atom, query_model.query_model.PropertyPath):
            orig_atom = orig_atom.getSingle()
        inverse_property = False # are we looking for a ^predicate
        if isinstance(orig_atom, query_model.query_model.PropertyPath.PathElt):
            inverse_property, orig_atom = orig_atom.inverse, orig_atom.pathPrimary
        if orig_atom in alias_map:
            alias = random.choice(alias_map[orig_atom])
            if self.show_queries:
                print()
                print(Colors.HIGHLIGHT+'Using alias: "{}"'.format(
                    alias)+Colors.ENDC)
        elif orig_atom.startswith('"') or orig_atom.startswith("'"):
            print("atom to complete is a literal, use it as its alias directly")
            orig_atom = re.sub(r"\^\^<[^\"]+", "", orig_atom)[:-1]
            orig_atom = re.sub(r"@[a-z]+$", "", orig_atom)
            orig_atom = orig_atom[1:-1]
            print("Using alias " + orig_atom)
            alias = orig_atom
            is_literal = True
        elif self.completion_mode != "predicate":
            print("subjects or objects without rdfs:label have to be typed manually (this includes literals")
            print("manually typing ", orig_atom)
            return orig_atom
        else:
            alias = orig_atom
            # always use the part after the prefix as alias
            if ':' in orig_atom and orig_atom[0] != "<":
                _, alias = alias.split(':')
            print(Colors.HIGHLIGHT + 'Using alias: "{}"'.format(
                alias) + Colors.ENDC)
            never_use_exact_regex = True

        # the first entry gets the new token bit.
        # The finish_log_record machinery sets it to False then on copying
        self.eval.current_record.is_new_token = True
        for i in count():
            if self.forced_typing_steps is None:
                prefix_pos = self.minimum_key_numbers + i * TYPING_STEP
            else:
                if i >= len(self.forced_typing_steps):
                    break
                prefix_pos = self.forced_typing_steps[i]

            use_exact_regex = prefix_pos > len(alias)
            # for the output, report how many letters we actually wanted to type
            self.eval.current_record.prefix_length = prefix_pos
            prefix_pos = min(prefix_pos, len(alias))
            prefix = alias[:prefix_pos]

            self.eval.current_record.prefix = prefix
            self.eval.current_record.alias = alias


            print(Colors.HIGHLIGHT +
                  'Typing: "{}"\n'.format(prefix) +
                  Colors.ENDC, end="")
            if len(prefix) > 0 and comp_query_something_typed is not None:
                query = deepcopy(comp_query_something_typed)
            else:
                query = deepcopy(comp_query)
            relevant_templates = [self.qlever_fallback_templates] if self.eval.current_record.is_warmup else [self.comp_templates]
            if self.use_agnostic_backup:
                relevant_templates.append(self.agnostic_comp_templates)
            if self.certify_fallback and not self.eval.current_record.is_warmup:
                relevant_templates.append(self.comp_templates)
            for idx, (el, templ) in enumerate(zip(query, relevant_templates)):
                replacements = dict()
                for name, repl in templ.items() :
                    replacements[name.upper()] = repl
                conditions = dict()
                conditions["CURRENT_WORD_EMPTY"] = True
                replacements["CURRENT_WORD"] = ""
                conditions["PREFIX_COMPLETE"] = use_exact_regex and not never_use_exact_regex
                if conditions["PREFIX_COMPLETE"]:
                    self.eval.current_record.prefix += "$"  # make it visible in the evaluation
                if prefix:
                    #TODO<joka921> how to deal with wdno:P40 etc.
                    search_str = prefix
                    # No need to search for '"' (added above for aliases) or '<'
                    if len(prefix) > 1 or (len(prefix) == 1 and prefix[0] not in "<\""):
                        conditions["CURRENT_WORD_EMPTY"] = False
                        replacements["CURRENT_WORD"] = SparqlCompleter.prefix_regex(
                            self.backend.is_qlever,
                            search_str) if not use_exact_regex else search_str


                replacements["CANDIDATE"] = "ql:autocomplete_certify_candidate"
                el = template_handling.evaluate_conditions(el, conditions, True)
                el = template_handling.substitute(el, replacements, True)
                global_prefixes = ""
                if "global-prefixes" in templ:
                    global_prefixes = templ["global-prefixes"]

                sparql = global_prefixes + '\n' + el
                if self.show_queries:
                    print('Completion query:')
                    print()
                    print(sparql)
                    print()
                query[idx] = query_model.parse_sparql(sparql, True)

            all_have_failed = self._complete_atom_with_query(
                current, orig_atom, query, inverse_property, only_use_agnostic, is_literal=is_literal)



            # Need to look for the parent which has a name
            named_parent = orig  # type: GraphPattern
            while not isinstance(named_parent, NamedQuery):
                assert named_parent.parent
                named_parent = named_parent.parent

            completion_query_name = \
                named_parent.name+'-completion-' + \
                str(self.completion_query_counter)
            self.eval.current_record.completion_query_name = \
                completion_query_name
            # Now write the eval record with all the informaiton
            positions = self.eval.current_record.num_completion_moves
            # Check if we are done
            was_last_step = False
            if self.forced_typing_steps is not None:
                was_last_step = i == len(self.forced_typing_steps) - 1
            elif len(prefix) >= len(alias):
                print("Typed the complete alias {} but could not finish completion!".format(alias))
                # Done completing
                was_last_step = True
            elif all_have_failed is None :
                # this signals an error and will count as failure
                was_last_step = True
            elif min (positions) < INSTANT_VIEWABLE :
                self.eval.steps_until_completion_map[len(prefix)] = self.eval.steps_until_completion_map.get(len(prefix), 0) + 1
                was_last_step = True
            if was_last_step:
                self.eval.current_record.is_last_query_of_token = True
            self.eval.finish_log_record()
            if was_last_step:
                return


    def _complete_subject(self, current: GroupGraphPattern,
                          orig: GroupGraphPattern,
                          completion_pos: CompletionPos,
                          alias_map: Dict[str, List[str]]) \
            -> Optional[str]:


        self.completion_mode = "subject"
        """
        Complete at a subject completion point i.e. missing subject
        """
        orig_pattern = orig[completion_pos.pattern_idx]
        orig_atom = orig_pattern[completion_pos.atom_idx]

        if not self.backend.is_qlever and self.mode == "sensitive":
            return orig_atom

        # completing a subject with no context, always use the  qlever-backend
        comp_query = self._completion_query(
            'subject', True)

        # enable the use of the %PREFIXES% template, expand it to an empty string
        replacements = {"PREFIXES" : ""}
        for idx, el in enumerate(comp_query):
            comp_query[idx] = template_handling.substitute(el, replacements, False)
        # we never have context for subjects, so this is always a warmup
        self.eval.current_record.is_warmup = True
        self.eval.current_record.has_context = False
        return self._complete_atom_with_queries(current, orig,
                                                orig_atom,
                                                comp_query,
                                                alias_map)


    def _subject_has_context(self, current, pattern):
        seeds = set()
        seeds.add(pattern[0])

        context = GroupGraphPattern()
        self.extend_with_context(context, current,
                                        seeds, self.replace_predicates)

        context = context.to_sparql()
        # strip the leading and trailing {}
        context = context[1:-1]
        return not context.isspace() and context

    def _complete_predicate(self, current: GroupGraphPattern,
                            orig: GroupGraphPattern,
                            completion_pos: CompletionPos,
                            alias_map: Dict[str, List[str]],
                            force_fixed_subject=False) \
            -> Optional[str]:
        """
        Complete at a predicate completion point i.e. missing predicate
        """

        # reinstate the default
        self.eval.current_record.is_warmup = False
        self.completion_mode = "predicate"
        pattern = current[completion_pos.pattern_idx]
        orig_pattern = orig[completion_pos.pattern_idx]
        orig_atom = orig_pattern[completion_pos.atom_idx]



        # completing a predicate
        seeds = set()  # type: Set[str]
        if pattern[2] and pattern[2] != '□':  # object constraint
            raise RuntimeError("Completions where the subject and object are given and the predicate is searched for are currently unsupported")
            #comp_query = self._completion_query(
            #    'predicate-subject-object')
            #comp_query.replace('?qlc_anchor_object', pattern[2])
            #seeds.add(pattern[2])
        else:  # no object constraint
            pass
        # Set subject of ql:has-predicate
        seeds.add(pattern[0])
        replacements = {"CURRENT_SUBJECT" : pattern[0]}
        replacements["PREFIXES"] = orig.prologue().to_sparql()
        conditions = {"CURRENT_SUBJECT_VARIABLE" : pattern[0].startswith("?")}
        context = GroupGraphPattern()
        #TODO<joka921> if we don't have any replace predicates, we do not mess with
        # prologues for the context
        #context.parent = orig # this is semantically wrong but the context needs the prologue
        #TODO<joka921> we have to deal with the "replacements better suited for" stuff
        self.extend_with_context(context, current,
                                            seeds, self.replace_predicates)

        context = context.to_sparql()
        # strip the leading and trailing {}
        context = context[1:-1]
        replacements["CONNECTED_TRIPLES"] = context
        conditions["CONNECTED_TRIPLES_EMPTY"] = context.isspace() or not context
        conditions["CONNECTED_TRIPLES_EMTPY_OR_CURRENT_SUBJECT_FIXED"] = conditions["CONNECTED_TRIPLES_EMPTY"] or (not conditions["CURRENT_SUBJECT_VARIABLE"])
        conditions["CURRENT_SUBJECT_FIXED"] = not conditions["CURRENT_SUBJECT_VARIABLE"] or force_fixed_subject
        conditions["NOT_CONNECTED_TRIPLES_EMPTY_AND_CURRENT_SUBJECT_VARIABLE"] = not conditions["CONNECTED_TRIPLES_EMTPY_OR_CURRENT_SUBJECT_FIXED"]
        conditions["CONNECTED_TRIPLES_EMPTY_AND_CURRENT_SUBJECT_VARIABLE"] = conditions["CONNECTED_TRIPLES_EMPTY"] and conditions["CURRENT_SUBJECT_VARIABLE"]

        fixed_subject = conditions["CURRENT_SUBJECT_FIXED"]
        self.eval.current_record.has_context = not conditions["CONNECTED_TRIPLES_EMPTY"] or fixed_subject

        if isinstance(orig_atom, PropertyPath):
            els = []
            for alt in orig_atom.alternatives:
                new_orig = deepcopy(orig)
                new_orig[completion_pos.pattern_idx][completion_pos.atom_idx] = alt
                els.append(self._complete_predicate(current, new_orig, completion_pos, alias_map, fixed_subject))
                if els[-1] is None:
                    return None
            return PropertyPath(els)

        elif isinstance(orig_atom, PropertyPath.PathSequence):
            seqs = []
            for seq in orig_atom.els:
                if seq.inverse:
                    print("Autocompletion for inverse paths is currently not supported, directly typing:", seq.to_sparql())
                    seqs.append(seq)
                    continue
                new_current = deepcopy(current)
                new_orig = deepcopy(orig)
                new_completion_pos = deepcopy(completion_pos)
                if seqs :
                    seqsCpy = deepcopy(seqs)
                    # make sure that the partial path is non-empty (qlever needs this)
                    seqsCpy[-1] = seqsCpy[-1].to_non_empty_path() if all(x.can_be_empty() for x in seqsCpy) else seqsCpy[-1]
                    pathSoFar = PropertyPath([PropertyPath.PathSequence(seqsCpy)])
                    new_current.insert(completion_pos.pattern_idx, [pattern[0], pathSoFar, "?qleverautocomplintermediate"])
                    new_orig.insert(completion_pos.pattern_idx, [pattern[0], pathSoFar, "?qleverautocomplintermediate"])
                    new_completion_pos.pattern_idx += 1
                    new_current[new_completion_pos.pattern_idx] = ["?qleverautocomplintermediate", '□', '']
                    new_orig[new_completion_pos.pattern_idx] = ["?qleverautocomplintermediate", seq.remove_modifiers(), new_orig[new_completion_pos.pattern_idx][2]]
                else:
                    # first element of the path
                    new_current[new_completion_pos.pattern_idx] = [pattern[0], '□', '']
                    X = new_orig[new_completion_pos.pattern_idx]
                    new_orig[new_completion_pos.pattern_idx] = [X[0], seq.remove_modifiers(), X[2]]

                try:
                    res = self._complete_predicate(new_current, new_orig, new_completion_pos, alias_map, fixed_subject)
                except BackendError:
                    res = None
                if res is None:
                    self.eval.num_failed += 1
                    completed_atom = new_orig[new_completion_pos.pattern_idx][new_completion_pos.atom_idx]
                    self.eval.num_completion_moves = self.eval.num_manual_moves
                    self.eval.stuck_targets.append(str(completed_atom))
                    print(Colors.FAIL + 'Completion of {} failed.'.format(
                        completed_atom) + Colors.ENDC)
                seqs.append(seq) # simulate typing the path modifier
            return PropertyPath.PathSequence(seqs)


        else:
            if conditions["CURRENT_SUBJECT_VARIABLE"] and conditions["CONNECTED_TRIPLES_EMPTY"]:
                self.eval.current_record.is_warmup = True
            else:
                # the recursion tree above us might already have set this to True
                self.eval.current_record.is_warmup = False
            if self.eval.current_record.is_warmup and not self.backend.is_qlever:
                return orig_atom
            comp_query = self._completion_query('predicate', self.eval.current_record.is_warmup)

            if not self.eval.current_record.is_warmup and self.certify_fallback:
                comp_query.append(self._completion_query('predicate-certify', False, True)[0])

            for idx, el in enumerate(comp_query):
                el = template_handling.evaluate_conditions(el, conditions)
                comp_query[idx] = template_handling.substitute(el, replacements)
            return self._complete_atom_with_queries(current, orig,
                                                orig_atom,
                                                comp_query,
                                                alias_map)

    def _complete_object(self, current: GroupGraphPattern,
                         orig: GroupGraphPattern,
                         completion_pos: CompletionPos,
                         alias_map: Dict[str, List[str]]) \
            -> Optional[str]:
        self.completion_mode = "object"
        """
        Complete at an object completion point i.e. missing object
        """
        pattern = current[completion_pos.pattern_idx]
        orig_pattern = orig[completion_pos.pattern_idx]
        orig_atom = orig_pattern[completion_pos.atom_idx]
        # reinstate the default
        self.eval.current_record.is_warmup = False

        # completing an object
        seeds = set()  # type: Set[str]
        if pattern[0] and pattern[0] != '□':  # subject constraint
            pass
        else:  # no subject constraint
            raise RuntimeError("The object can currently only complete when the subject and predicate are already known")
            #comp_query = self._completion_query(
            #    current,
            #    'object-no-subject-predicate')
            #comp_query_aliases = self._completion_query(
            #    current,
            #    'object-no-subject-predicate-aliases')
        # Set predicate of template
        seeds.add(pattern[0])
        seeds.add(pattern[1])
        replacements = {"CURRENT_SUBJECT" : pattern[0]}
        conditions = {"CURRENT_SUBJECT_VARIABLE" : pattern[0].startswith("?")}
        #predicate = orig.full_iri(pattern[1].getSingle() if isinstance(pattern[1], query_model.query_model.PropertyPath) else pattern[1])
        predicate = pattern[1]
        if (isinstance(predicate, PropertyPath)):
            if self.superset_predicates:
                predicate = predicate.get_superset()
            predicate = predicate.to_sparql()
        conditions["CURRENT_PREDICATE_VARIABLE"] = predicate.startswith("?")
        if predicate in self.replace_predicates:
            predicate = self.replace_predicates[predicate]

        replacements["CURRENT_PREDICATE"] = predicate



        replacements["PREFIXES"] = orig.prologue().to_sparql()
        context = GroupGraphPattern()
        context.parent = orig # the context has to access the prologue of orig.
        #TODO<joka921> we have to deal with the "replacements better suited for" stuff
        self.extend_with_context(context, current,
                                            seeds, self.replace_predicates)


        context = context.to_sparql()
        # strip the leading and trailing {}
        context = context[1:-1]
        replacements["CONNECTED_TRIPLES"] = context
        conditions["CONNECTED_TRIPLES_EMPTY"] = context.isspace() or not context
        self.eval.current_record.has_context = not conditions["CURRENT_SUBJECT_VARIABLE"] or not (conditions["CONNECTED_TRIPLES_EMPTY"] and conditions["CURRENT_PREDICATE_VARIABLE"])
        conditions["CONNECTED_TRIPLES_EMTPY_OR_CURRENT_SUBJECT_FIXED"] = conditions["CONNECTED_TRIPLES_EMPTY"] or (not conditions["CURRENT_SUBJECT_VARIABLE"])
        conditions["NOT_CONNECTED_TRIPLES_EMPTY_AND_CURRENT_SUBJECT_VARIABLE"] = not conditions["CONNECTED_TRIPLES_EMTPY_OR_CURRENT_SUBJECT_FIXED"]

        if conditions["CURRENT_SUBJECT_VARIABLE"] and conditions["CONNECTED_TRIPLES_EMPTY"] and replacements["CURRENT_PREDICATE"] in self.warmup_predicates:
            self.eval.current_record.is_warmup = True

        if conditions["CONNECTED_TRIPLES_EMPTY"] and conditions["CURRENT_SUBJECT_VARIABLE"] and conditions["CURRENT_PREDICATE_VARIABLE"]:
            # use the completion query for unconstrained subjects
            if self.eval.current_record.is_warmup and not self.backend.is_qlever:
                return orig_atom
            comp_query_something_typed = self._completion_query('subject', True)
            comp_query_nothing_typed = self._completion_query('subject', True)
            self.eval.current_record.is_warmup = True
        else:
            if self.eval.current_record.is_warmup and not self.backend.is_qlever:
                return orig_atom
            comp_query_nothing_typed = self._completion_query('object', self.eval.current_record.is_warmup)
            comp_query_something_typed = deepcopy(comp_query_nothing_typed)
            comp_query_something_typed[0] = self._completion_query('object-long-prefix', self.eval.current_record.is_warmup, True)[0]
            if comp_query_something_typed is None or comp_query_something_typed[0] is None:
                comp_query_something_typed = self._completion_query('object', self.eval.current_record.is_warmup)

            if self.certify_fallback:
                cert = self._completion_query("object-certify", False, True)
                comp_query_something_typed.append(cert[0])
                comp_query_nothing_typed.append(cert[0])

        def create_transitive_type_pred():
            a = PropertyPath.PathElt("wdt:P31")
            b = PropertyPath.PathElt("wdt:P279", False, "*")
            return PropertyPath([PropertyPath.PathSequence([a, b])])

        # disable this for now, as we don't want special stuff.
        def is_pred_special(pred):
            #return pred == create_transitive_type_pred()
            return False

        if (not context or context.isspace()) and conditions["CURRENT_SUBJECT_VARIABLE"] and is_pred_special(predicate):
            print("Using the warmed up query for instance_of/subclass* without context")
            self.eval.current_record.is_warmup = True
            if self.eval.current_record.is_warmup and not self.backend.is_qlever:
                return orig_atom
            if "transitive-type" in self.qlever_fallback_templates:
                comp_query = self._completion_query("transitive-type", True)
            else:
                print("No special template found, complete as  a normal object")
        for c in [comp_query_nothing_typed, comp_query_something_typed]:
            for idx, el in enumerate(c):
                el = template_handling.evaluate_conditions(el, conditions)
                c[idx] = template_handling.substitute(el, replacements)

        res =  self._complete_atom_with_queries(current, orig,
                                                orig_atom,
                                                comp_query_nothing_typed,
                                                alias_map, comp_query_something_typed)

        return res

    def _complete_triple(self,
                         current, orig: GroupGraphPattern,
                         pattern, orig_pattern: List[str],
                         pattern_idx: int,
                         alias_map: Dict[str, List[str]]) \
            -> Optional[List[str]]:
        """
        Complete a triple where any one of subject, predicate or object may be
        a completion marker
        """
        assert(len(pattern) == 3)




        for atom_idx, atom in enumerate(pattern):
            if not atom or atom != '□':
                continue
            completed_atom = None
            self.eval.current_record.pattern = deepcopy(pattern)
            self.eval.current_record.orig_pattern = \
                deepcopy(orig_pattern)
            self.eval.current_record.pattern_idx = pattern_idx
            self.eval.num_completion_atoms += 1

            orig_atom = orig_pattern[atom_idx]
            if isinstance(orig_atom, query_model.query_model.PropertyPath) and orig_atom.isSingle():
                orig_atom = orig_atom.getSingle()

            dummy_query_for_variable=False
            skip=False
            if not isinstance(orig_atom, query_model.query_model.PropertyPath) and orig_atom.startswith('?'):
                if self.query_variables:
                    # we are sending a single dummy query for the variables
                    # We need an alias for the atom which is empty so we don't start typing
                    #TODO: does it work this way?
                    dummy_query_for_variable = True
                    alias_map[orig_atom] = [""]
                    print("Sending a dummy query for variable " + orig_atom)
                else:
                    self.eval.num_completion_moves += len(orig_atom)
                    completed_atom = orig_atom
                    skip=True

            if not skip:
                if atom_idx == 0:
                    self.eval.needed_completion += 1
                    try:
                        completed_atom = \
                            self._complete_subject(
                                current, orig,
                                CompletionPos(pattern_idx, atom_idx),
                                alias_map)
                        #Todo: is this qlever-friendly?
                    except BackendError:
                        completed_atom = None
                elif atom_idx == 1:
                    self.eval.needed_completion += 1
                    completed_atom = \
                        self._complete_predicate(
                            current, orig,
                            CompletionPos(pattern_idx, atom_idx),
                            alias_map)
                elif atom_idx == 2:
                    self.eval.needed_completion += 1
                    completed_atom = \
                        self._complete_object(
                            current, orig,
                            CompletionPos(pattern_idx, atom_idx),
                            alias_map)
                else:
                    raise ValueError('Triple with > 3 atoms')

            if dummy_query_for_variable:
                alias_map.pop(orig_atom)
                completed_atom=orig_atom
                print("Completed variable to {} using a dummy query".format(completed_atom))

            self.eval.num_manual_moves += len(orig_atom)
            new_triple = deepcopy(pattern)
            if completed_atom is None:
                self.eval.num_failed += 1
                completed_atom = orig_pattern[atom_idx]
                self.eval.num_completion_moves = self.eval.num_manual_moves
                self.eval.stuck_targets.append(str(completed_atom))
                print(Colors.FAIL+'Completion of {} failed.'.format(
                    completed_atom)+Colors.ENDC)
            else:
                print('completed to', completed_atom)
            new_triple[atom_idx] = completed_atom
            if atom_idx < 2:
                new_triple[atom_idx + 1] = '□'
            return new_triple
        return None

    def _completion_step(self, current: GroupGraphPattern,
                         orig: GroupGraphPattern,
                         alias_map: Dict[str, List[str]]) -> GroupGraphPattern:
        """
        Do a single completion step on current to bring it closer to matching
        orig
        """
        current = deepcopy(current)
        self.eval.num_completion_steps += 1
        if not SparqlCompleter.has_completion_point(current) and \
                len(current.patterns()) < len(orig.patterns()):
            # currently we only complete left to right, top to bottom
            # but this is not necessary we could complete in the middle
            # or right to left
            current.append(['□', '', ''])
        if self.show_queries:
            print('-'*80)
            print('completing:\n')
            print(current.to_sparql().replace(
                '□', Colors.HIGHLIGHT+'□'+Colors.ENDC))
            print()
        for pattern_idx, pattern in enumerate(current):
            orig_pattern = orig[pattern_idx]
            if pattern == orig_pattern:
                # Already completed
                continue
            if isinstance(orig_pattern, OptionalPattern):
                if pattern == orig_pattern:
                    # Already completed
                    continue
                group_pattern = GroupGraphPattern([pattern])
                group_pattern.parent = current
                current[pattern_idx] = group_pattern
                new_pattern = self._complete_group_graph_pattern(
                    group_pattern,
                    orig_pattern.pattern,
                    alias_map)
                if not new_pattern:
                    return current
                optional_pattern = OptionalPattern(new_pattern)
                current[pattern_idx] = optional_pattern
                if self.show_queries:
                    print('completed to', optional_pattern.to_sparql())
                return current
            if isinstance(orig_pattern, MinusGraphPattern):
                if pattern == orig_pattern:
                    # Already completed
                    continue
                group = GroupGraphPattern()
                group.parent = current
                # temporarily register the first half of the union within the parent to make
                # the context propagation work
                saved = current[pattern_idx]
                current[pattern_idx] = group
                new_pattern = self._complete_group_graph_pattern(
                    group,
                    orig_pattern.pattern,
                    alias_map)
                if not new_pattern:
                    current[pattern_idx] = saved # do we need this?
                    return current

                minus_pattern = MinusGraphPattern(new_pattern)
                current[pattern_idx] = minus_pattern
                if self.show_queries:
                    print('completed to', minus_pattern.to_sparql())
                return current
            if isinstance(orig_pattern, UnionPattern):
                if pattern == orig_pattern:
                    # Already completed
                    continue
                lhs_group = GroupGraphPattern()
                lhs_group.parent = current
                # temporarily register the first half of the union within the parent to make
                # the context propagation work
                saved = current[pattern_idx]
                current[pattern_idx] = lhs_group
                lhs = self._complete_group_graph_pattern(
                    lhs_group,
                    orig_pattern.lhs,
                    alias_map)
                if not lhs:
                    current[pattern_idx] = saved # do we need this?
                    return current
                rhs_group = GroupGraphPattern()
                rhs_group.parent = current
                current[pattern_idx] = rhs_group
                rhs = self._complete_group_graph_pattern(
                    rhs_group,
                    orig_pattern.rhs,
                    alias_map)
                if not rhs:
                    current[pattern_idx] = saved # do we need this?
                    return current
                union_pattern = UnionPattern(lhs, rhs)
                current[pattern_idx] = union_pattern
                print('completed to', union_pattern.to_sparql())
                return current
            if isinstance(orig_pattern, list):
                # Looking at a Triple
                if isinstance(pattern, MinusGraphPattern):
                    raise NotImplementedError("Wrong type, breakpoint")
                if '□' not in pattern:
                    continue

                if self.exclude_labels and isinstance(orig_pattern[1], PropertyPath) and orig_pattern[1].isSingle() and (orig_pattern[1].getSingle()=="rdfs:label" or orig_pattern[1].getSingle()=="fb:type.object.name"):
                    print("Excluding triple with rdfs:label predicate, typing it directly")
                    current[pattern_idx]=orig_pattern
                    continue

                #TODO proper flag for this
                if orig_pattern[0].startswith("?") and not orig_pattern[2].startswith(
                        "?") and self.reverse_fixed_object_triples\
                        and not self._subject_has_context(current, orig_pattern) :
                    if not isinstance(orig_pattern[1], str):
                        # might be a variable...
                        orig_pattern[1].reverse()
                    orig_pattern.reverse()
                    print("reversed triple {}".format(orig_pattern))
                    orig.completion_triple_is_reversed = True
                new_triple = \
                    self._complete_triple(
                        current, orig,
                        pattern, orig_pattern,
                        pattern_idx,
                        alias_map)
                current[pattern_idx] = new_triple
                if new_triple == orig_pattern and orig.completion_triple_is_reversed:
                    if not isinstance(orig_pattern[1], str):
                        # might be a variable...
                        orig_pattern[1].reverse()
                        new_triple[1].reverse()
                    orig_pattern.reverse()
                    new_triple.reverse()
                    orig.completion_triple_is_reversed = False

                return current
            if isinstance(orig_pattern, SelectQuery):
                innerCurrent = SelectQuery(orig_pattern.owned_prologue,
                                      orig_pattern.distinct,
                                      orig_pattern.selectors,
                                      GroupGraphPattern(),
                                      orig_pattern.group,
                                      orig_pattern.order,
                                      orig_pattern.limit,
                                      orig_pattern.parent)
                innerCurrent.where = self._complete_group_graph_pattern(innerCurrent.where,
                        orig_pattern.where, alias_map)
                current[pattern_idx] = innerCurrent
                return current

            if isinstance(orig_pattern, ValuesPattern):
                print(Colors.HIGHLIGHT +
                      'Pattern to complete is not a triple!' +
                      Colors.ENDC)

            if isinstance(orig_pattern, list) and (not pattern or '□' not in pattern):
                continue
            print(Colors.HIGHLIGHT +
                  'Pattern to complete is not a triple and will thus not be completed (FILTER, VALUE, BIND clauses e.g.)' +
                  Colors.ENDC)
            completed = orig[pattern_idx]
            completed_sparql = GraphPattern.pattern_to_sparql(completed)
            if self.show_queries:
                print('completed to', completed_sparql)
            current[pattern_idx] = completed
            return current
        # Nothing was completed further
        return current

    def _get_entities(self, query: GroupGraphPattern):
        """
        Creates a map from the shortnames used in the actual query to
        their aliases that a user would be able to guess
        """
        aliases = set()  # type: Set[str]
        for pattern in query.patterns():
            if isinstance(pattern, UnionPattern):
                aliases.update(self._get_entities(pattern.lhs))
                aliases.update(self._get_entities(pattern.rhs))
            if isinstance(pattern, OptionalPattern) or isinstance(pattern, MinusGraphPattern):
                aliases.update(self._get_entities(pattern.pattern))

            if isinstance(pattern, SelectQuery):
                # subqueries
                aliases.update(self._get_entities(pattern.where))
            elif isinstance(pattern, list):
                for external_atom in pattern:
                    atoms = {}
                    if isinstance(external_atom, query_model.query_model.PropertyPath):
                        atoms = external_atom.all_elements()
                    else:
                        atoms = {external_atom}
                    for atom in atoms:
                        if not atom.startswith('?') and not atom.startswith('"'):
                            if not self.backend.is_qlever and self.backend.is_blazegraph and atom.startswith("fbe:"):
                                atom_full_iri = "<file:/local/data/blazegraph/freebase-easy/" + re.sub(r'\\(.)', r'\1', atom[4:]) + '>'
                            else:
                                atom_full_iri = query.full_iri(atom)
                            if atom_full_iri in self.replace_predicates:
                                atom = self.replace_predicates[atom_full_iri]
                            aliases.add((atom, atom_full_iri))
        return aliases


    def _create_alias_map(self, query: GroupGraphPattern) \
            -> Dict[str, List[str]]:
        """
        Creates a map from the shortnames used in the actual query to
        their aliases that a user would be able to guess
        """
        alias_map = {}  # type: Dict[str, List[str]]
        if False:
        #if not self.backend.is_qlever and not self.backend.is_blazegraph:
            alias_query_template = self._completion_query('lookup-aliases', True)[0]
            global_prefixes = self._completion_query("global-prefixes", True)[0]
        else:
            alias_query_template = self._completion_query('lookup-aliases', False)[0]
            global_prefixes = self._completion_query("global-prefixes", False)[0]
        if not alias_query_template:
            return alias_map
        global_prefixes = "" if not global_prefixes else global_prefixes
        alias_query_template = global_prefixes + '\n' + alias_query_template
        entities = list(self._get_entities(query))
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []

            for atom, atom_full_iri in entities:
                    alias_query = deepcopy(alias_query_template)
                    # allow the old ?qlc_complete and the new %LOOKUP_ENTITY%
                    alias_query = alias_query.replace('?qlc_complete', atom_full_iri)
                    replacements = {"LOOKUP_ENTITY": atom_full_iri}
                    if False:
                    #if not self.backend.is_qlever and not self.backend.is_blazegraph:
                        completion_items = self.qlever_fallback_templates
                    else:
                        completion_items = self.comp_templates

                    for name, repl in completion_items.items():
                        replacements[name.upper()] = repl
                    alias_query = template_handling.substitute(alias_query, replacements, True)
                    if self.show_queries:
                        print('Alias query:\n'+alias_query)
                    alias_query = query_model.parse_sparql(alias_query, True)
                    if self.show_queries:
                        print('Alias query:\n'+alias_query.to_sparql())

                    sp = alias_query.to_sparql()
                    if False:
                    #if not self.backend.is_qlever and not self.backend.is_blazegraph:
                        # virtuoso is for some reason missing some labels
                        futures.append(executor.submit(lambda: self.qlever_fallback_backed.query(sp, timeout=30)))
                    else:
                        futures.append(executor.submit(lambda :self.backend.query(sp, timeout = 30)))
                    #res, _, _ = self.backend.query(sp, timeout = 30) #avoid  timeout in alias query
            for i, fut in enumerate(futures):
                    atom, atom_full_iri = entities[i]
                    res, _, _ = fut.result()
                    if res:
                        aliases = \
                            [alias[0][1:alias[0].rfind('"')]
                             for alias in res]
                        print('aliases for:', atom_full_iri, 'are', aliases)
                        if any(x.isascii() for x in aliases):
                            aliases = list(filter(lambda x: x.isascii(), aliases))
                        alias_map[atom] = aliases
                    else:
                        print('No aliases found for:', atom_full_iri)
        if self.show_queries:
            print()
        return alias_map

    def _complete_group_graph_pattern(self, current, orig: GroupGraphPattern,
                                      alias_map: Dict[str, List[str]]) \
            -> Optional[GroupGraphPattern]:
        """
        Complete a GroupGraphPattern
        """
        while current != orig:
            try:
                new = \
                    self._completion_step(current,
                                          orig,
                                          alias_map)
            except BackendError:
                return None
            if current == new:
                print(Colors.FAIL+'Got stuck on target'+Colors.ENDC)
                if self.show_queries:
                    print(Colors.FAIL)
                    print(orig.to_sparql())
                print("Current completion status is: \n" + new.to_sparql() +
                      '\n'+Colors.ENDC)
                return None
            current = new
        return current

    def complete(self, orig: NamedQuery, alias_map=None):
        """
        Tries to generate the given original SPARQL query by using
        auto completions
        """
        if self.show_queries:
            print('='*80)
            print('Target SPARQL:')
            print()
            print(orig.to_sparql())
            print()
            print(Colors.OKGREEN+"Retrieving aliases"+Colors.ENDC)
        if alias_map is None:
            alias_map = self._create_alias_map(orig.where)
        current = SelectQuery(orig.prologue(),
                              False,
                              orig.selectors,
                              GroupGraphPattern(),
                              orig.group,
                              orig.order,
                              orig.limit)
        # Reset per query completion query counter
        self.completion_query_counter = 0
        self.eval.current_record.query_name = orig.name

        return self._complete_group_graph_pattern(
            current.where, orig.where, alias_map)
